# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Incident Features."""
from datetime import datetime
from typing import Dict, List, Optional, Union
from uuid import UUID, uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import _azs_api_result_to_df, _build_sent_data, get_http_timeout

__version__ = VERSION
__author__ = "Pete Bryan"


class SentinelIncidentsMixin:
    """Mixin class for Sentinel Incidents feature integrations."""

    def get_incident(
        self,
        incident: str,
        entities: bool = False,
        alerts: bool = False,
        comments: bool = False,
        bookmarks: bool = False,
    ) -> pd.DataFrame:
        """
        Get details on a specific incident.

        Parameters
        ----------
        incident : str
            Incident ID GUID.
        entities : bool, optional
            If True include all entities in the response. Default is False.
        alerts : bool, optional
            If True include all alerts in the response. Default is False.
        comments: bool, optional
             If True include all comments in the response. Default is False.
        bookmarks: bool, optional
             If True include all bookmarks in the response. Default is False.

        Returns
        -------
        pd.DataFrame
            Table containing incident details.

        Raises
        ------
        CloudError
            If incident could not be retrieved.

        """
        incident_id = self._get_incident_id(incident)
        incident_url = self.sent_urls["incidents"] + f"/{incident_id}"  # type: ignore
        response = self._get_items(incident_url)  # type: ignore
        if response.status_code != 200:
            raise CloudError(response=response)

        incident_df = _azs_api_result_to_df(response)

        if entities:
            incident_df["Entities"] = [self.get_entities(incident_id)]

        if alerts:
            incident_df["Alerts"] = [self.get_incident_alerts(incident_id)]

        if comments:
            incident_df["Bookmarks"] = [self.get_incident_comments(incident_id)]

        if bookmarks:
            incident_df["Bookmarks"] = [self.get_incident_bookmarks(incident_id)]

        return incident_df

    def get_entities(self, incident: str) -> list:
        """
        Get the entities from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name .

        Returns
        -------
        list
            A list of entities.

        """
        self.check_connected()  # type: ignore
        incident_id = self._get_incident_id(incident)
        entities_url = self.sent_urls["incidents"] + f"/{incident_id}/entities"  # type: ignore
        ent_parameters = {"api-version": "2021-04-01"}
        ents = httpx.post(
            entities_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=ent_parameters,
            timeout=get_http_timeout(),
        )
        return (
            [(ent["kind"], ent["properties"]) for ent in ents.json()["entities"]]
            if ents.status_code == 200
            else []
        )

    def get_incident_alerts(self, incident: str) -> list:
        """
        Get the alerts from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name.

        Returns
        -------
        list
            A list of alerts.

        """
        self.check_connected()  # type: ignore
        incident_id = self._get_incident_id(incident)
        alerts_url = self.sent_urls["incidents"] + f"/{incident_id}/alerts"  # type: ignore
        alerts_parameters = {"api-version": "2021-04-01"}
        alerts_resp = httpx.post(
            alerts_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=alerts_parameters,
            timeout=get_http_timeout(),
        )
        return (
            [
                {
                    "ID": alrts["properties"]["systemAlertId"],
                    "Name": alrts["properties"]["alertDisplayName"],
                }
                for alrts in alerts_resp.json()["value"]
            ]
            if alerts_resp.status_code == 200
            else []
        )

    def get_incident_comments(self, incident: str) -> list:
        """
        Get the comments from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name.

        Returns
        -------
        list
            A list of comments.

        """
        incident_id = self._get_incident_id(incident)
        comments_url = self.sent_urls["incidents"] + f"/{incident_id}/comments"  # type: ignore
        comments_response = self._get_items(comments_url, "2021-04-01")  # type: ignore
        comment_details = comments_response.json()
        return (
            [
                {
                    "Message": comment["properties"]["message"],
                    "Author": comment["properties"]["author"]["name"],
                }
                for comment in comment_details["value"]
            ]
            if comments_response.status_code == 200
            else []
        )

    def get_incident_bookmarks(self, incident: str) -> list:
        """
        Get the comments from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or name.

        Returns
        -------
        list
            A list of bookmarks.

        """
        bookmarks_list = []
        incident_id = self._get_incident_id(incident)
        relations_url = self.sent_urls["incidents"] + f"/{incident_id}/relations"  # type: ignore
        relations_response = self._get_items(relations_url, "2021-04-01")  # type: ignore
        if relations_response.status_code == 200 and relations_response.json()["value"]:
            for relationship in relations_response.json()["value"]:
                if (
                    relationship["properties"]["relatedResourceType"]
                    == "Microsoft.SecurityInsights/Bookmarks"
                ):
                    bkmark_id = relationship["properties"]["relatedResourceName"]
                    bookmarks_df = self.list_bookmarks()  # type: ignore
                    bookmark = bookmarks_df[bookmarks_df["name"] == bkmark_id].iloc[0]
                    bookmarks_list.append(
                        {
                            "Bookmark ID": bkmark_id,
                            "Bookmark Title": bookmark["properties.displayName"],
                        }
                    )

        return bookmarks_list

    def update_incident(
        self,
        incident_id: str,
        update_items: dict,
    ):
        """
        Update properties of an incident.

        Parameters
        ----------
        incident_id : str
            Incident ID GUID.
        update_items : dict
            Dictionary of properties to update and their values.
            https://docs.microsoft.com/rest/api/securityinsights/
            stable/incidents/create-or-update

        Raises
        ------
        CloudError
            If incident could not be updated.

        """
        self.check_connected()  # type: ignore
        incident_dets = self.get_incident(incident_id)
        incident_url = self.sent_urls["incidents"] + f"/{incident_id}"  # type: ignore
        params = {"api-version": "2020-01-01"}
        if "title" not in update_items.keys():
            update_items["title"] = incident_dets.iloc[0]["properties.title"]
        if "status" not in update_items.keys():
            update_items["status"] = incident_dets.iloc[0]["properties.status"]
        data = _build_sent_data(
            update_items, props=True, etag=incident_dets.iloc[0]["etag"]
        )
        response = httpx.put(
            incident_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        print("Incident updated.")
        return response.json().get("name")

    def create_incident(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        title: str,
        severity: str,
        status: str = "New",
        description: Optional[str] = None,
        first_activity_time: Optional[datetime] = None,
        last_activity_time: Optional[datetime] = None,
        labels: Optional[List] = None,
        bookmarks: Optional[List] = None,
    ) -> Optional[str]:
        """
        Create a Sentinel Incident.

        Parameters
        ----------
        title : str
            The title of the incident to create
        severity : str
            The severity to assign the incident, options are:
               Informational, Low, Medium, High
        status : str, optional
            The status to assign the incident, by default "New"
            Options are: New, Active, Closed
        description : str, optional
            A description of the incident, by default None
        first_activity_time : datetime, optional
            The start time of the incident activity, by default None
        last_activity_time : datetime, optional
            The end time of the incident activity, by default None
        labels : List, optional
            Any labels to apply to the incident, by default None
        bookmarks : List, optional
            A list of bookmark GUIDS you want to associate with the incident

        Returns
        -------
        Optional[str]
            The name/ID of the incident.

        Raises
        ------
        CloudError
            If the API returns an error

        """
        self.check_connected()  # type: ignore
        incident_id = uuid4()
        incident_url = self.sent_urls["incidents"] + f"/{incident_id}"  # type: ignore
        params = {"api-version": "2020-01-01"}
        data_items: Dict[str, Union[str, List]] = {
            "title": title,
            "severity": severity.capitalize(),
            "status": status.capitalize(),
        }
        if description:
            data_items["description"] = description
        if labels:
            labels = [{"labelName": lab, "labelType": "User"} for lab in labels]
            data_items["labels"] = labels
        if first_activity_time:
            data_items["firstActivityTimeUtc"] = first_activity_time.isoformat()
        if last_activity_time:
            data_items["lastActivityTimeUtc"] = last_activity_time.isoformat()
        data = _build_sent_data(data_items, props=True)
        response = httpx.put(
            incident_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        if bookmarks:
            for mark in bookmarks:
                relation_id = uuid4()
                bookmark_id = self._get_bookmark_id(mark)  # type: ignore
                mark_res_id = self.sent_urls["bookmarks"] + f"/{bookmark_id}"  # type: ignore
                relations_url = incident_url + f"/relations/{relation_id}"
                bkmark_data_items = {"relatedResourceId": mark_res_id}
                data = _build_sent_data(bkmark_data_items, props=True)
                params = {"api-version": "2021-04-01"}
                response = httpx.put(
                    relations_url,
                    headers=get_api_headers(self._token),  # type: ignore
                    params=params,
                    content=str(data),
                    timeout=get_http_timeout(),
                )
        print("Incident created.")
        return response.json().get("name")

    def _get_incident_id(self, incident: str) -> str:
        """
        Get an incident ID.

        Parameters
        ----------
        incident : str
            An incident identifier

        Returns
        -------
        str
            The Incident GUID

        Raises
        ------
        MsticpyUserError
            If incident can't be found or multiple matching incidents found.

        """
        try:
            UUID(incident)
            return incident
        except ValueError as incident_name:
            incidents = self.list_incidents()
            filtered_incidents = incidents[
                incidents["properties.title"].str.contains(incident)
            ]
            if len(filtered_incidents) > 1:
                display(filtered_incidents[["name", "properties.title"]])
                raise MsticpyUserError(
                    "More than one incident found, please specify by GUID"
                ) from incident_name
            if (
                not isinstance(filtered_incidents, pd.DataFrame)
                or filtered_incidents.empty
            ):
                raise MsticpyUserError(
                    f"Incident {incident} not found"
                ) from incident_name
            return filtered_incidents["name"].iloc[0]

    def post_comment(
        self,
        incident_id: str,
        comment: str,
    ):
        """
        Write a comment for an incident.

        Parameters
        ----------
        incident_id : str
            Incident ID GUID.
        comment : str
            Comment message to post.

        Raises
        ------
        CloudError
            If message could not be posted.

        """
        self.check_connected()  # type: ignore
        comment_url = (
            self.sent_urls["incidents"] + f"/{incident_id}/comments/{uuid4()}"  # type: ignore
        )
        params = {"api-version": "2020-01-01"}
        data = _build_sent_data({"message": comment})
        response = httpx.put(
            comment_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        print("Comment posted.")
        return response.json().get("name")

    def add_bookmark_to_incident(self, incident: str, bookmark: str):
        """
        Add a bookmark to an incident.

        Parameters
        ----------
        incident : str
            Either an incident name or an incident GUID
        bookmark : str
            Either a bookmark name or bookmark GUID

        Raises
        ------
        CloudError
            If API returns error

        """
        self.check_connected()  # type: ignore
        incident_id = self._get_incident_id(incident)
        incident_url = self.sent_urls["incidents"] + f"/{incident_id}"  # type: ignore
        bookmark_id = self._get_bookmark_id(bookmark)  # type: ignore
        mark_res_id = self.sent_urls["bookmarks"] + f"/{bookmark_id}"  # type: ignore
        relations_id = uuid4()
        bookmark_url = incident_url + f"/relations/{relations_id}"
        bkmark_data_items = {
            "relatedResourceId": mark_res_id.split(self.base_url)[1]  # type: ignore
        }
        data = _build_sent_data(bkmark_data_items, props=True)
        params = {"api-version": "2021-04-01"}
        response = httpx.put(
            bookmark_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        print("Bookmark added to incident.")
        return response.json().get("name")

    def list_incidents(self, params: Optional[dict] = None) -> pd.DataFrame:
        """
        Get a list of incident for a Sentinel workspace.

        Parameters
        ----------
        params : Optional[dict], optional
            Additional parameters to pass to the API call, by default None

        Returns
        -------
        pd.DataFrame
            A table of incidents.

        Raises
        ------
        CloudError
            If incidents could not be retrieved.

        """
        if params is None:
            params = {"$top": 50}
        return self._list_items(item_type="incidents", params=params)  # type: ignore

    get_incidents = list_incidents



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
AbuseIPDB Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from typing import Any, Dict, Tuple

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .result_severity import ResultSeverity
from .ti_http_provider import HttpTIProvider

__version__ = VERSION
__author__ = "Rubén Revuelta Briz"


@export
class AbuseIPDB(HttpTIProvider):
    """AbuseIPDB Lookup."""

    PROVIDER_NAME = "AbuseIPDB"

    _BASE_URL = "https://api.abuseipdb.com/"

    _QUERIES = {
        "ipv4": APILookupParams(
            path="api/v2/check",
            params={"ipAddress": "{observable}"},
            headers={"Key": "{AuthKey}", "Accept": "application/json"},
        )
    }

    # aliases
    _QUERIES["ipv6"] = _QUERIES["ipv4"]

    def parse_results(self, response: Dict) -> Tuple[bool, ResultSeverity, Any]:
        """Return the details of the response."""
        if self._failed_response(response) or not isinstance(
            response["RawResult"], dict
        ):
            return False, ResultSeverity.information, "Not found."

        data = response["RawResult"]["data"]
        result_dict = {
            "countryCode": data.get("countryCode", None),
            "usage": data.get("usageType", None),
            "isp": data.get("isp", None),
            "domain": data.get("domain", None),
            "hostNames": data.get("hostnames", None),
            "lastReportedAt": data.get("lastReportedAt", None),
        }

        score = data.get("abuseConfidenceScore", None)

        if score == 0:
            result_severity = ResultSeverity.information
        elif score <= 50:
            result_severity = ResultSeverity.warning
        elif score > 50:
            result_severity = ResultSeverity.high

        return (True, result_severity, result_dict)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy authentication and secretesmodules.

This is sub-package authentication and configuration
modules and classes.

Some of the modules and classes here:

- azure_auth - az_connect function
- secret_settings - handles Azure Key Vault and keyring-stored secrets

"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# Author: Thomas Roccia - @fr0gger_
# --------------------------------------------------------------------------
"""MalwareBazaar TI Provider."""
import json
from enum import Enum
from typing import Set

import httpx
import pandas as pd
from pandas import json_normalize

from ..._version import VERSION
from ...common.provider_settings import get_provider_settings

__version__ = VERSION
__author__ = "Thomas Roccia | @fr0gger_"

_BASE_URL = "https://mb-api.abuse.ch/api/v1/"

_QUERY_OBJECTS_MAPPINGS = {
    "hash": {"query": "get_info", "hash": "observable"},
    "tag": {"query": "get_taginfo", "tag": "observable", "limit": "limit"},
    "signature": {"query": "get_siginfo", "signature": "observable", "limit": "limit"},
    "filetype": {"query": "get_file_type", "file_type": "observable", "limit": "limit"},
    "clamav": {"query": "get_clamavinfo", "clamav": "observable", "limit": "limit"},
    "imphash": {"query": "get_imphash", "imphash": "observable", "limit": "limit"},
    "dhash": {"query": "get_dhash_icon", "dhash_icon": "observable", "limit": "limit"},
    "yara": {"query": "get_yarainfo", "yara_rule": "observable", "limit": "limit"},
    "tlsh": {"query": "get_tlsh", "tlsh": "observable", "limit": "limit"},
    "telfhash": {"query": "get_telfhash", "telfhash": "observable", "limit": "limit"},
    "gimphash": {"query": "get_gimphash", "gimphash": "observable", "limit": "limit"},
    "issuerinfo": {"query": "get_issuerinfo", "issuer_cn": "observable"},
    "subjectinfo": {"query": "get_subjectinfo", "subject_cn": "observable"},
    "certificate": {"query": "get_certificate", "serial_number": "observable"},
}


class MBEntityType(Enum):
    """MBEntityType: Enum class for MalwareBazaar entity types."""

    HASH = "hash"
    TAG = "tag"
    SIGNATURE = "signature"
    FILETYPE = "filetype"
    CLAMAV = "clamav"
    IMPHASH = "imphash"
    DHASH = "dhash"
    YARA = "yara"
    TLSH = "tlsh"
    TELFHASH = "telfhash"
    CODESIGNISSUER = "issuerinfo"
    CODESIGNSUBJECT = "subjectinfo"
    CODESIGNSN = "certificate"
    GIMPHASH = "gimphash"


class MBlookup:
    """MBlookup Python Class wrapper for MalwareBazaar API."""

    _SUPPORTED_MB_TYPES: Set[MBEntityType] = {
        MBEntityType.HASH,
        MBEntityType.TAG,
        MBEntityType.SIGNATURE,
        MBEntityType.FILETYPE,
        MBEntityType.CLAMAV,
        MBEntityType.IMPHASH,
        MBEntityType.DHASH,
        MBEntityType.YARA,
        MBEntityType.TLSH,
        MBEntityType.TELFHASH,
        MBEntityType.CODESIGNISSUER,
        MBEntityType.CODESIGNSUBJECT,
        MBEntityType.CODESIGNSN,
        MBEntityType.GIMPHASH,
    }

    def __init__(self, mb_key=None):
        """Init function to get the API key if necessary."""
        self.mb_key = mb_key  # or_get_mb_api_key()

    def _make_mb_request(self, data):
        """Request to the malware bazaar api."""
        try:
            res = httpx.post(_BASE_URL, data=data, timeout=None)
            res_data = json.loads(res.text)
            if res_data["query_status"] == "ok":
                return json_normalize(res_data["data"])
            return res["query_status"]
        except httpx.ConnectError as err:
            return err

    def lookup_ioc(self, observable: str, mb_type: str, limit=10) -> pd.DataFrame:
        """
        Lookup for IOC in MalwareBazaar.

        Parameters
        ----------
        observable : str
            The observable to lookup. It can be a hash, a signature
        mb_type : str
            The type of the observable. It can be a hash, a signature (refer to MBEntityType).
        limit : int, optional
            The number of results to return, default is 100 or 50 in some cases, by default 10

        Returns
        -------
        pd.DataFrame
            The results of the lookup.

        Raises
        ------
        KeyError
            If invalid IoC type is provided.

        """
        if MBEntityType(mb_type) not in self._SUPPORTED_MB_TYPES:
            raise KeyError(
                f"Property type {mb_type} not supported",
                "Valid types are",
                ", ".join(x.value for x in MBEntityType.__members__.values()),
            )

        query_parameters = _build_query_string(observable, mb_type, limit)

        return self._make_mb_request(query_parameters)

    def download_sample(self, sha2: str):
        """Download specified sample from MB."""
        return self._make_mb_request({"query": "get_file", "sha256_hash": sha2})

    def get_recent(self, selector: str):
        """
        Get the recent MB additions.

        Parameters
        ----------
        selector : str
            Get the latest sample from the last 60 min.

        Returns
        -------
        pd.DataFrame
           The results of the latest addition.

        """
        return self._make_mb_request({"query": "get_recent", "selector": selector})

    def get_cscb(self):
        """Query Code Signing Certificate Blocklist (CSCB)."""
        return self._make_mb_request({"query": "get_cscb"})


def _get_mb_api_key():
    """Retrieve the MB key from settings."""
    prov_settings = get_provider_settings("TIProviders")
    mb_settings = prov_settings.get("MalwareBazaar")
    if mb_settings:
        return mb_settings.args.get("AuthKey")
    return None


def _build_query_string(observable, mb_type, limit):
    """Build the query string for the MB API."""
    query_items = _QUERY_OBJECTS_MAPPINGS[mb_type]
    if "limit" in query_items:
        query_items["limit"] = limit
    for key, value in query_items.items():
        if value == "observable":
            query_items[key] = observable
    return query_items



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Utilties."""
from collections import Counter
from typing import Any, Dict, List, Optional

import httpx
import pandas as pd
from azure.common.exceptions import CloudError

from ..._version import VERSION
from ...auth.azure_auth_core import AzureCloudConfig
from ...common.exceptions import MsticpyAzureConfigError, MsticpyAzureConnectionError
from ...common.pkg_config import get_http_timeout
from ...common.wsconfig import WorkspaceConfig
from .azure_data import get_api_headers

__version__ = VERSION
__author__ = "Pete Bryan"

_PATH_MAPPING = {
    "ops_path": "/providers/Microsoft.SecurityInsights/operations",
    "alert_rules": "/providers/Microsoft.SecurityInsights/alertRules",
    "ss_path": "/savedSearches",
    "bookmarks": "/providers/Microsoft.SecurityInsights/bookmarks",
    "incidents": "/providers/Microsoft.SecurityInsights/incidents",
    "data_connectors": "/providers/Microsoft.SecurityInsights/dataConnectors",
    "watchlists": "/providers/Microsoft.SecurityInsights/watchlists",
    "alert_template": "/providers/Microsoft.SecurityInsights/alertRuleTemplates",
    "search": "/tables",
    "ti": "/providers/Microsoft.SecurityInsights/threatIntelligence/main",
    "dynamic_summary": "/providers/Microsoft.SecurityInsights/dynamicSummaries",
}


class SentinelUtilsMixin:
    """Mixin class for Sentinel core feature integrations."""

    def _get_items(self, url: str, params: Optional[dict] = None) -> httpx.Response:
        """Get items from the API."""
        self.check_connected()  # type: ignore
        if params is None:
            params = {"api-version": "2020-01-01"}
        return httpx.get(
            url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            timeout=get_http_timeout(),
        )

    def _list_items(
        self,
        item_type: str,
        api_version: str = "2020-01-01",
        appendix: Optional[str] = None,
        next_follow: bool = False,
        params: Optional[Dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """
        Return lists of core resources from APIs.

        Parameters
        ----------
        item_type : str
            The type of resource you want to list.
        api_version: str, optional
            The API version to use, by default '2020-01-01'
        appendix: str, optional
            Any appendix that needs adding to the URI, default is None
        next_follow: bool, optional
            If True, follow the nextLink to get all results, by default False
        params: Dict, optional
            Any additional parameters to pass to the API call, by default None
        Returns
        -------
        pd.DataFrame
            A DataFrame containing the requested items.

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        item_url = self.url + _PATH_MAPPING[item_type]  # type: ignore
        if appendix:
            item_url = item_url + appendix
        if params is None:
            params = {}
        params["api-version"] = api_version
        response = self._get_items(item_url, params)
        if response.status_code == 200:
            results_df = _azs_api_result_to_df(response)
        else:
            raise CloudError(response=response)
        j_resp = response.json()
        results = [results_df]
        # If nextLink in response, go get that data as well
        if next_follow:
            i = 0
            # Limit to 5 nextLinks to prevent infinite loop
            while "nextLink" in j_resp and i < 5:
                next_url = j_resp["nextLink"]
                next_response = self._get_items(next_url, params)
                next_results_df = _azs_api_result_to_df(next_response)
                results.append(next_results_df)
                j_resp = next_response.json()
                i += 1
        return pd.concat(results)

    def _check_config(self, items: List, workspace_name: Optional[str] = None) -> Dict:
        """
        Get parameters from default config files.

        Parameters
        ----------
        items : List
            The items to get from the config.
        workspace_name : str
            The workspace name supplied by the user.

        Returns
        -------
        Dict
            The config items.

        """
        config_items = {}
        if not self.workspace_config:  # type: ignore
            self.workspace_config = WorkspaceConfig(workspace_name)  # type: ignore
        for item in items:
            if item in self.workspace_config:  # type: ignore
                config_items[item] = self.workspace_config[item]  # type: ignore
            else:
                raise MsticpyAzureConfigError(
                    f"No {item} available in config for workspace {workspace_name}."
                )

        return config_items

    def _build_sent_res_id(
        self,
        sub_id: Optional[str] = None,
        res_grp: Optional[str] = None,
        ws_name: Optional[str] = None,
    ) -> str:
        """
        Build a resource ID.

        Parameters
        ----------
        sub_id : str, optional
            Subscription ID to use, by default None
        res_grp : str, optional
            Resource Group name to use, by default None
        ws_name : str, optional
            Workspace name to user, by default None

        Returns
        -------
        str
            The formatted resource ID.

        """
        if not sub_id or not res_grp or not ws_name:
            config = self._check_config(
                workspace_name=ws_name,
                items=[
                    WorkspaceConfig.CONF_SUB_ID,
                    WorkspaceConfig.CONF_RES_GROUP,
                    WorkspaceConfig.CONF_WS_NAME,
                ],
            )
            sub_id = config[WorkspaceConfig.CONF_SUB_ID]
            res_grp = config[WorkspaceConfig.CONF_RES_GROUP]
            ws_name = config[WorkspaceConfig.CONF_WS_NAME]
        return "".join(
            [
                f"/subscriptions/{sub_id}/resourcegroups/{res_grp}",
                f"/providers/Microsoft.OperationalInsights/workspaces/{ws_name}",
            ]
        )

    def _build_sent_paths(self, res_id: str, base_url: Optional[str] = None) -> str:
        """
        Build an API URL from an Azure resource ID.

        Parameters
        ----------
        res_id : str
            An Azure resource ID.
        base_url : str, optional
            The base URL of the Azure cloud to connect to.
            Defaults to resource manager for configured cloud.
            If no cloud configuration, defaults to resource manager
            endpoint for public cloud.

        Returns
        -------
        str
            A URI to that resource.

        """
        if not base_url:
            base_url = AzureCloudConfig(self.cloud).resource_manager  # type: ignore
        res_info = {
            "subscription_id": res_id.split("/")[2],
            "resource_group": res_id.split("/")[4],
            "workspace_name": res_id.split("/")[-1],
        }

        return "".join(
            [
                f"{base_url}subscriptions/{res_info['subscription_id']}",
                f"/resourceGroups/{res_info['resource_group']}",
                "/providers/Microsoft.OperationalInsights/workspaces"
                f"/{res_info['workspace_name']}",
            ]
        )

    def check_connected(self):
        """Check that Sentinel workspace is connected."""
        if not self.connected:  # type: ignore
            raise MsticpyAzureConnectionError(
                "Not connected to Sentinel, ensure you run `.connect` before calling functions."
            )


def _azs_api_result_to_df(response: httpx.Response) -> pd.DataFrame:
    """
    Convert API response to a Pandas dataframe.

    Parameters
    ----------
    response : httpx.Response
        A response object from an Azure REST API call.

    Returns
    -------
    pd.DataFrame
        The API response as a Pandas dataframe.

    Raises
    ------
    ValueError
        If the response is not valid JSON.

    """
    j_resp = response.json()
    if response.status_code != 200 or not j_resp:
        raise ValueError("No valid JSON result in response")
    if "value" in j_resp:
        j_resp = j_resp["value"]
    return pd.json_normalize(j_resp)


def _build_sent_data(items: dict, props: bool = False, **kwargs) -> dict:
    """
    Build request data body from items.

    Parameters
    ----------
    items : dict
        A set pf items to be formated in the request body.
    props: bool, optional
        Whether all items are to be built as properities. Default is false.

    Returns
    -------
    dict
        The request body formatted for the API.

    """
    data_body = {"properties": {}}  # type: Dict[str, Dict[str, str]]
    for key in items:
        if key in ["severity", "status", "title", "message", "searchResults"] or props:
            data_body["properties"].update({key: items[key]})  # type:ignore
        else:
            data_body[key] = items[key]
    if "etag" in kwargs:
        data_body["etag"] = kwargs.get("etag")  # type:ignore
    return data_body


def validate_res_id(res_id):
    """Validate a Resource ID String and fix if needed."""
    valid = _validator(res_id)
    if not valid:
        res_id = _fix_res_id(res_id)
        valid = _validator(res_id)
    if not valid:
        raise MsticpyAzureConfigError("The Resource ID provided is not valid.")

    return res_id


def parse_resource_id(res_id: str) -> Dict[str, Any]:
    """Extract components from workspace resource ID."""
    res_id_parts = res_id.split("/")
    return {
        "subscription_id": res_id_parts[1] if len(res_id_parts) > 1 else None,
        "resource_group": res_id_parts[3] if len(res_id_parts) > 3 else None,
        "workspace_name": res_id_parts[7] if len(res_id_parts) > 7 else None,
    }


def _validator(res_id):
    """Check Resource ID string matches pattern expected."""
    counts = Counter(res_id)
    return bool(
        res_id.startswith("/") and counts["/"] == 8 and not res_id.endswith("/")
    )


def _fix_res_id(res_id):
    """Try to fix common issues with Resource ID string."""
    if res_id.startswith("https:"):
        res_id = "/".join(res_id.split("/")[5:])
    if not res_id.startswith("/"):
        res_id = "/" + res_id
    if res_id.endswith("/"):
        res_id = res_id[:-1]
    counts = Counter(res_id)
    if counts["/"] > 8:
        res_id = "/".join(res_id.split("/")[:9])
    return res_id



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Analytics Features."""
from typing import Optional
from uuid import UUID, uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import _build_sent_data, get_http_timeout

__version__ = VERSION
__author__ = "Pete Bryan"


class SentinelHuntingMixin:
    """Mixin class for Sentinel Hunting feature integrations."""

    def list_hunting_queries(self) -> pd.DataFrame:
        """
        Return all custom hunting queries in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        saved_query_df = self._list_items(  # type: ignore
            item_type="ss_path", api_version="2020-08-01"
        )
        return saved_query_df[
            saved_query_df["properties.category"] == "Hunting Queries"
        ]

    get_hunting_queries = list_hunting_queries

    def list_saved_queries(self) -> pd.DataFrame:
        """
        Return all saved queries in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        saved_query_df = self._list_items(  # type: ignore
            item_type="ss_path", api_version="2020-08-01"
        )
        return saved_query_df

    get_hunting_queries = list_hunting_queries


class SentinelAnalyticsMixin:
    """Mixin class for Sentinel Analytics feature integrations."""

    def list_alert_rules(self) -> pd.DataFrame:
        """
        Return all Microsoft Sentinel alert rules for a workspace.

        Returns
        -------
        pd.DataFrame
            A table of the workspace's alert rules.

        """
        return self._list_items(  # type: ignore
            item_type="alert_rules", api_version="2022-11-01"
        )

    def _get_template_id(
        self,
        template: str,
    ) -> str:
        """
        Get an analytic template ID.

        Parameters
        ----------
        template : str
            Template ID or Name
        res_id : str
            Sentinel workspace to get template from

        Returns
        -------
        str
            Template ID

        Raises
        ------
        MsticpyUserError
            If template not found or multiple templates found.\

        """
        try:
            UUID(template)
            return template
        except ValueError as template_name:
            templates = self.list_analytic_templates()
            template_details = templates[
                templates["properties.displayName"].str.contains(template)
            ]
            if len(template_details) > 1:
                display(template_details[["name", "properties.displayName"]])
                raise MsticpyUserError(
                    "More than one template found, please specify by GUID"
                ) from template_name
            if not isinstance(template_details, pd.DataFrame) or template_details.empty:
                raise MsticpyUserError(
                    f"Template {template_details} not found"
                ) from template_name
            return template_details["name"].iloc[0]

    def create_analytic_rule(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        template: str = None,
        name: str = None,
        enabled: bool = True,
        query: str = None,
        query_frequency: str = "PT5H",
        query_period: str = "PT5H",
        severity: str = "Medium",
        suppression_duration: str = "PT1H",
        suppression_enabled: bool = False,
        trigger_operator: str = "GreaterThan",
        trigger_threshold: int = 0,
        description: str = None,
        tactics: list = None,
    ) -> Optional[str]:
        """
        Create a Sentinel Analytics Rule.

        Parameters
        ----------
        template : str, optional
            The GUID or name of a templated to create the analytic from, by default None
        name : str, optional
            The name to give the analytic, by default None
        enabled : bool, optional
            Whether you want the analytic to be enabled once deployed, by default True
        query : str, optional
            The KQL query string to use in the anlaytic, by default None
        query_frequency : str, optional
            How often the query should run in ISO8601 format, by default "PT5H"
        query_period : str, optional
            How far back the query should look in ISO8601 format, by default "PT5H"
        severity : str, optional
            The severity to raise incidents as, by default "Medium"
            Options are; Informational, Low, Medium, or High
        suppression_duration : str, optional
            How long to suppress duplicate alerts in ISO8601 format, by default "PT1H"
        suppression_enabled : bool, optional
            Whether you want to suppress duplicates, by default False
        trigger_operator : str, optional
            The operator for the trigger, by default "GreaterThan"
        trigger_threshold : int, optional
            The threshold of events required to create the incident, by default 0
        description : str, optional
            A description of the analytic, by default None
        tactics : list, optional
            A list of MITRE ATT&CK tactics related to the analytic, by default None

        Returns
        -------
        Optional[str]
            The name/ID of the analytic rule.

        Raises
        ------
        MsticpyUserError
            If template provided isn't found.
        CloudError
            If the API returns an error.

        """
        self.check_connected()  # type: ignore
        if template:
            template_id = self._get_template_id(template)
            templates = self.list_analytic_templates()
            template_details = templates[templates["name"] == template_id].iloc[0]
            name = template_details["properties.displayName"]
            query = template_details["properties.query"]
            query_frequency = template_details["properties.queryFrequency"]
            query_period = template_details["properties.queryPeriod"]
            severity = template_details["properties.severity"]
            trigger_operator = template_details["properties.triggerOperator"]
            trigger_threshold = template_details["properties.triggerThreshold"]
            description = template_details["properties.description"]
            tactics = (
                template_details["properties.tactics"]
                if not pd.isna(template_details["properties.tactics"])
                else []
            )

        if not tactics:
            tactics = []

        if not name:
            raise MsticpyUserError(
                "Please specify either a template ID or analytic details."
            )

        rule_id = uuid4()
        analytic_url = self.sent_urls["alert_rules"] + f"/{rule_id}"  # type: ignore
        data_items = {
            "displayName": name,
            "query": query,
            "queryFrequency": query_frequency,
            "queryPeriod": query_period,
            "severity": severity,
            "suppressionDuration": suppression_duration,
            "suppressionEnabled": str(suppression_enabled).lower(),
            "triggerOperator": trigger_operator,
            "triggerThreshold": trigger_threshold,
            "description": description,
            "tactics": tactics,
            "enabled": str(enabled).lower(),
        }
        data = _build_sent_data(data_items, props=True)
        data["kind"] = "Scheduled"
        params = {"api-version": "2020-01-01"}
        response = httpx.put(
            analytic_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code != 201:
            raise CloudError(response=response)
        print("Analytic Created.")
        return response.json().get("name")

    def _get_analytic_id(self, analytic: str) -> str:
        """
        Get the GUID of an analytic rule.

        Parameters
        ----------
        analytic : str
            The GUID or name of the analytic

        Returns
        -------
        str
            The analytic GUID

        Raises
        ------
        MsticpyUserError
            If analytic not found or multiple matching analytics found

        """
        try:
            UUID(analytic)
            return analytic
        except ValueError as analytic_name:
            analytics = self.list_analytic_rules()
            analytic_details = analytics[
                analytics["properties.displayName"].str.contains(analytic)
            ]
            if len(analytic_details) > 1:
                display(analytic_details[["name", "properties.displayName"]])
                raise MsticpyUserError(
                    "More than one analytic found, please specify by GUID"
                ) from analytic_name
            if not isinstance(analytic_details, pd.DataFrame) or analytic_details.empty:
                raise MsticpyUserError(
                    f"Analytic {analytic_details} not found"
                ) from analytic_name
            return analytic_details["name"].iloc[0]

    def delete_analytic_rule(
        self,
        analytic_rule: str,
    ):
        """
        Delete a deployed Analytic rule from a Sentinel workspace.

        Parameters
        ----------
        analytic_rule : str
            The GUID or name of the analytic.

        Raises
        ------
        CloudError
            If the API returns an error.

        """
        self.check_connected()  # type: ignore
        analytic_id = self._get_analytic_id(analytic_rule)
        analytic_url = self.sent_urls["alert_rules"] + f"/{analytic_id}"  # type: ignore
        params = {"api-version": "2020-01-01"}
        response = httpx.delete(
            analytic_url,
            headers=get_api_headers(self._token),  # type: ignore
            params=params,
            timeout=get_http_timeout(),
        )
        if response.status_code != 200:
            raise CloudError(response=response)
        print("Analytic Deleted.")

    def list_analytic_templates(self) -> pd.DataFrame:
        """
        List Analytic Templates.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the analytics templates

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        return self._list_items(item_type="alert_template")  # type: ignore

    get_alert_rules = list_alert_rules
    list_analytic_rules = list_alert_rules
    get_analytic_rules = list_alert_rules



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""VirusTotal V3 Subpackage."""
import contextlib

from ..._version import VERSION

# pylint: disable=unused-import
# flake8: noqa: F401
VT3_AVAILABLE = False
with contextlib.suppress(ImportError):
    import nest_asyncio
    import vt
    import vt_graph_api

    VT3_AVAILABLE = True

if VT3_AVAILABLE:
    from ...vis.vtobject_browser import VTObjectBrowser
    from . import vtlookupv3
    from .vtfile_behavior import VTFileBehavior
    from .vtlookupv3 import (
        VT_API_NOT_FOUND,
        MsticpyVTNoDataError,
        VTEntityType,
        VTLookupV3,
        VTObjectProperties,
    )

__version__ = VERSION
__author__ = "Ian Hellen"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPy plugin loader.

To load MSTICPy plugins you must have modules with classes
defined from one of the supported plugin types.

To specify the locations to look for plugins specify
the folder(s) as absolute paths or relative to the current
directory

msticpyconfig.yaml
------------------
PluginFolders:
  - tests/testdata/dp_plugins
  - tests/testdata/ti_plugins
  - tests/testdata/mixed_plugins

"""


import contextlib
import sys
from importlib import import_module
from inspect import getmembers, isabstract, isclass
from pathlib import Path
from types import ModuleType
from typing import Iterable, NamedTuple, Optional, Union
from warnings import warn

from .._version import VERSION
from ..common.pkg_config import get_config
from ..context.contextlookup import ContextLookup
from ..context.contextproviders.context_provider_base import ContextProvider
from ..context.tilookup import TILookup
from ..context.tiproviders.ti_provider_base import TIProvider
from ..data import drivers
from ..data.drivers import DriverBase

__version__ = VERSION
__author__ = "Ian Hellen"

_PLUGIN_KEY = "PluginFolders"


class PluginReg(NamedTuple):
    """Plugin registration tuple."""

    reg_dest: Union[type, ModuleType]  # class or module containing CUSTOM_PROVIDERS
    name_property: Optional[str]  # Custom name(s) for provider


# This dictionary maps the class of the plugin to
# the module or class where the CUSTOM_PROVIDERS dictionary
# for that provider type is defined.
# This dictionary maps the class of the plugin to
# the module or class where the CUSTOM_PROVIDERS dictionary
# for that provider type is defined.
_PLUGIN_TYPES = {
    DriverBase: PluginReg(drivers, "DATA_ENVIRONMENTS"),
    TIProvider: PluginReg(TILookup, "PROVIDER_NAME"),
    ContextProvider: PluginReg(ContextLookup, "PROVIDER_NAME"),
}


def read_plugins(plugin_paths: Union[str, Iterable[str]]):
    """Load plugins from folders specified in msticpyconfig.yaml."""
    plugin_config = [plugin_paths] if isinstance(plugin_paths, str) else plugin_paths
    if not plugin_config:
        with contextlib.suppress(KeyError):
            plugin_config = get_config(_PLUGIN_KEY)
    if not plugin_config:
        return
    for plugin_path in plugin_config:
        load_plugins_from_path(plugin_path=plugin_path)


def load_plugins_from_path(plugin_path: Union[str, Path]):
    """Load all compatible plugins found in plugin_path."""
    sys.path.append(str(plugin_path))
    for module_file in Path(plugin_path).glob("*.py"):
        try:
            module = import_module(module_file.stem)
        except ImportError:
            warn(f"Unable to import plugin {module_file} from {plugin_path}")
        for name, obj in getmembers(module, isclass):
            if not isinstance(obj, type):
                continue
            if isabstract(obj):
                continue
            plugin_type, reg_object = _get_plugin_type(obj)
            if plugin_type:
                # if no specified registration, use the root class
                reg_dest = reg_object.reg_dest or plugin_type
                plugin_names = getattr(obj, reg_object.name_property, name)
                if not isinstance(plugin_names, (list, tuple)):
                    plugin_names = (plugin_names,)
                for plugin_name in plugin_names:
                    reg_dest.CUSTOM_PROVIDERS[plugin_name] = obj


def _get_plugin_type(plugin_class):
    """Return matching parent class for plugin_class."""
    return next(
        (
            (parent_class, plugin_reg)
            for parent_class, plugin_reg in _PLUGIN_TYPES.items()
            if issubclass(plugin_class, parent_class) and plugin_class != parent_class
        ),
        (None, None),
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Tor Exit Nodes Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""


import contextlib
from datetime import datetime, timezone
from threading import Lock
from typing import Any, Dict, Iterable, Tuple

import httpx
import pandas as pd

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ...common.utility import export, mp_ua_header
from ..lookup_result import LookupStatus
from .ti_provider_base import ResultSeverity, TIProvider

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class Tor(TIProvider):
    """Tor Exit Nodes Lookup."""

    _BASE_URL = "https://check.torproject.org/exit-addresses"
    _ALT_URL = (
        "https://raw.githubusercontent.com/SecOps-Institute/Tor-IP-Addresses"
        "/master/tor-exit-nodes.lst"
    )

    _QUERIES: Dict[str, Any] = {"ipv4": None, "ipv6": None}
    _nodelist: Dict[str, Dict[str, str]] = {}
    _last_cached = datetime.min
    _cache_lock = Lock()

    @classmethod
    def _check_and_get_nodelist(cls):
        """Pull down Tor exit node list and save to internal attribute."""
        if cls._cache_lock.locked():
            return
        now = datetime.now(timezone.utc)
        if not cls._nodelist or (now - cls._last_cached).days > 1:
            with contextlib.suppress(ConnectionError):
                resp = httpx.get(
                    cls._BASE_URL, timeout=get_http_timeout(), headers=mp_ua_header()
                )
                tor_raw_list = resp.content.decode()
                with cls._cache_lock:
                    cls._nodelist = dict(cls._tor_splitter(tor_raw_list))
                    cls._last_cached = datetime.now(timezone.utc)
        if not cls._nodelist:
            with contextlib.suppress(ConnectionError):
                resp = httpx.get(
                    cls._ALT_URL, timeout=get_http_timeout(), headers=mp_ua_header()
                )
                tor_raw_list = resp.content.decode()
                with cls._cache_lock:
                    node_dict = {"ExitNode": True, "LastStatus": now}
                    cls._nodelist = {
                        node: node_dict for node in tor_raw_list.split("\n")
                    }
                    cls._last_cached = datetime.now(timezone.utc)

    @staticmethod
    def _tor_splitter(node_list) -> Iterable[Tuple[str, Dict[str, str]]]:
        node_dict: Dict[str, str] = {}
        for line in node_list.split("\n"):
            if not line:
                continue
            fields = line.split(" ", 2)
            if fields[0] == "ExitNode":
                # new record so reset dict
                node_dict = {}
            node_dict[fields[0]] = fields[1] if len(fields) > 1 else None
            if fields[0] == "ExitAddress":
                # yield tuple
                yield fields[1], node_dict

    def lookup_ioc(
        self, ioc: str, ioc_type: str = None, query_type: str = None, **kwargs
    ) -> pd.DataFrame:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            The returned results.

        """
        if not self._nodelist:
            self._check_and_get_nodelist()
        result = self._check_ioc_type(
            ioc=ioc, ioc_type=ioc_type, query_subtype=query_type
        )

        result["Provider"] = kwargs.get("provider_name", self.__class__.__name__)
        result["Result"] = bool(self._nodelist)
        result["Reference"] = self._BASE_URL

        if result["Status"] and not bool(self._nodelist):
            result["Status"] = LookupStatus.QUERY_FAILED.value

        if result["Status"]:
            return pd.DataFrame([result])

        tor_node = self._nodelist.get(ioc)

        if tor_node:
            result["Severity"] = ResultSeverity.warning.name
            result["Details"] = {
                "NodeID": tor_node["ExitNode"],
                "LastStatus": tor_node["LastStatus"],
            }
            result["RawResult"] = tor_node
        else:
            result["Details"] = "Not found."
        return pd.DataFrame([result])

    def parse_results(self, response: Dict) -> Tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        return (True, ResultSeverity.information, None)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Jupyter and Python Tools for InfoSec.

-----------------------------------------------

Requires Python 3.8 or later.

Getting Started
---------------

To quickly import common modules into a notebook run:

>>> import msticpy as mp
>>> mp.init_notebook()

If not running in a notebook/IPython use
>>> mp.init_notebook(globals())

To see help on `init_notebook`:
>>> help(mp.init_notebook)

Search msticpy modules for a keyword:
>>> mp.search(keyword)

-----------------------------------------------

Full documentation is available at:
https://msticpy.readthedocs.io

GitHub repo:
https://github.com/microsoft/msticpy

Getting started with msticpy
https://msticpy.readthedocs.io/en/latest/getting_started/QuickStart.html

Configuration
https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html

Package structure
-----------------

- analysis - analysis functions
- auth - authentication and secrets management
- common - utility functions, common types, exceptions
- config - configuration tool
- data - queries, data access, context functions
- datamodel - entities and pivot functions
- init - package initialization
- nbtools - deprecated location
- nbwidgets - notebook widgets
- resources - data resource files
- transform - data transforms and decoding
- vis - visualizations

Configuration
-------------

Set MSTICPYCONFIG environment variable to point to the path
of your `msticpyconfig.yaml` file.

One frequent problem in bootstrapping MSTICPy is obtaining or
creating a valid `msticpyconfig.yaml`. This is needed for many
configuration settings such as Data providers,
MS Sentinel workspaces, Threat Intelligence (TI) providers,
Azure authentication, Key Vault settings and more.

If you are having trouble with this, you should check out the
following resources:

The basic structure of msticpyconfig.yaml

https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html

Using the msticpy settings editor to create or modify
msticpyconfig.yaml

https://msticpy.readthedocs.io/en/latest/getting_started/SettingsEditor.html

MS Sentinel and Azure ML resources:

Getting started notebook
https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/A%20Getting%20Started%20Guide%20For%20Azure%20Sentinel%20ML%20Notebooks.ipynb

Configuring your environment notebook
https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/ConfiguringNotebookEnvironment.ipynb

Initialization
--------------

Running the nbinit function performs a number of functions
that allow you start working in a notebook more quickly:

>>> import msticpy as mp
>>> mp.init_notebook()

This module performs several steps to initialize MSTICPy:

- imports a number of standard packages (e.g. pandas) into the notebook
- imports a number of modules and functions from msticpy
- checks the version of MSTICPy
- checks for presence and validates msticpyconfig.yaml
- imports entities and initialized pivot functions
- initializes pandas with the .mp and .mp_plot accessors that
  give you quick access to many MSTICPy functions from a pandas
  DataFrame
- creates some MSTICPy notebook magics.

In the Azure ML and Azure Synapse environments, some additional
initialization and checks are performed.

"""
import os
from typing import Iterable, Union

from . import nbwidgets  # noqa: F401

# flake8: noqa: F403
from ._version import VERSION
from .common import pkg_config as settings
from .common.check_version import check_version  # noqa: F401
from .common.utility import search_name as search  # noqa: F401
from .init.logging import set_logging_level, setup_logging  # noqa: F401
from .lazy_importer import lazy_import

__version__ = VERSION
__author__ = "Ian Hellen, Pete Bryan, Ashwin Patil"

refresh_config = settings.refresh_config
get_config = settings.get_config
setup_logging()

if not os.environ.get("KQLMAGIC_EXTRAS_REQUIRES"):
    os.environ["KQLMAGIC_EXTRAS_REQUIRES"] = "jupyter-basic"

_LAZY_IMPORTS = {
    "msticpy.auth.azure_auth.az_connect",
    "msticpy.common.timespan.TimeSpan",
    "msticpy.common.wsconfig.WorkspaceConfig",
    "msticpy.config.mp_config_edit.MpConfigEdit",
    "msticpy.config.mp_config_file.MpConfigFile",
    "msticpy.context.azure.MicrosoftSentinel",
    "msticpy.context.contextlookup.ContextLookup",
    "msticpy.context.geoip.GeoLiteLookup",
    "msticpy.context.tilookup.TILookup",
    "msticpy.data.QueryProvider",
    "msticpy.datamodel.entities",
    "msticpy.init.nbinit.current_providers",
    "msticpy.init.nbinit.init_notebook",
    "msticpy.init.nbinit.reset_ipython_exception_handler",
    "msticpy.init.pivot.Pivot",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)


def load_plugins(plugin_paths: Union[str, Iterable[str]]):
    """
    Load plugins from specified paths or configuration.

    Parameters
    ----------
    plugin_paths : Union[str, Iterable[str]]
        A path or collection of paths from which to
        load plugins. If not supplied, msticpyconfig is checked for
        a PluginFolders key and list of paths are read from there.

    Notes
    -----
    No attempt to load plugins is made if both parameter and
    configuration are empty.

    """
    # pylint: disable=import-outside-toplevel
    from .init.mp_plugins import read_plugins

    read_plugins(plugin_paths)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
ip_utils - IP Address functions.

Contains a series of functions required to manipulate and
enrich IP Address data to assist investigations.

Designed to support any data source containing IP address entity.

"""
import ipaddress
import re
import socket
import warnings
from functools import lru_cache
from time import sleep
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Set, Tuple, Union

import httpx
import pandas as pd
from bs4 import BeautifulSoup
from deprecated.sphinx import deprecated

from .._version import VERSION
from ..common.exceptions import MsticpyConnectionError, MsticpyException
from ..common.utility import arg_to_list, export
from ..datamodel.entities import GeoLocation, IpAddress

__version__ = VERSION
__author__ = "Ashwin Patil"


_REGISTRIES = {
    "arin": {
        "url": "http://rdap.arin.net/registry/ip/",
    },
    "ripencc": {
        "url": "http://rdap.db.ripe.net/ip/",
    },
    "apnic": {
        "url": "http://rdap.apnic.net/ip/",
    },
    "lacnic": {
        "url": "http://rdap.lacnic.net/rdap/ip/",
    },
    "afrinic": {
        "url": "http://rdap.afrinic.net/rdap/ip/",
    },
}

_POTAROO_ASNS_URL = "https://bgp.potaroo.net/cidr/autnums.html"


# Closure to cache ASN dictionary from Potaroo
def _fetch_asns() -> Callable[[], Dict[str, str]]:
    """Create closure for ASN fetching."""
    asns_dict: Dict[str, str] = {}

    def _get_asns_dict() -> Dict[str, str]:
        """Return or fetch and return ASN Soup."""
        nonlocal asns_dict  # noqa
        if not asns_dict:
            try:
                asns_resp = httpx.get(_POTAROO_ASNS_URL)
            except httpx.ConnectError as err:
                raise MsticpyConnectionError(
                    "Unable to get ASN details from potaroo.net"
                ) from err
            asns_soup = BeautifulSoup(asns_resp.content, features="lxml")
            asns_dict = {
                str(asn.next_element)
                .strip(): str(asn.next_element.next_element)
                .strip()
                for asn in asns_soup.find_all("a")
            }
        return asns_dict

    return _get_asns_dict


# Create the dictionary accessor from the fetch_asns wrapper
_ASNS_DICT = _fetch_asns()


@export  # noqa: MC0001
def convert_to_ip_entities(  # noqa: MC0001
    ip_str: Optional[str] = None,
    data: Optional[pd.DataFrame] = None,
    ip_col: Optional[str] = None,
    geo_lookup: bool = True,
) -> List[IpAddress]:  # noqa: MC0001
    """
    Take in an IP Address string and converts it to an IP Entity.

    Parameters
    ----------
    ip_str : str
        A string with a single IP Address or multiple addresses
        delimited by comma or space
    data : pd.DataFrame
        Use DataFrame as input
    ip_col : str
        Column containing IP addresses
    geo_lookup : bool
        If true, do geolocation lookup on IPs,
        by default, True

    Returns
    -------
    List
        The populated IP entities including address and geo-location

    Raises
    ------
    ValueError
        If neither ip_string or data/column provided as input

    """
    # locally imported to prevent cyclic import
    # pylint: disable=import-outside-toplevel, cyclic-import
    from .geoip import GeoLiteLookup

    geo_lite_lookup = GeoLiteLookup()

    ip_entities: List[IpAddress] = []
    all_ips: Set[str] = set()

    if ip_str:
        addrs = arg_to_list(ip_str)
    elif data is not None and ip_col:
        addrs = data[ip_col].values
    else:
        raise ValueError("Must specify either ip_str or data + ip_col parameters.")

    for addr in addrs:
        if isinstance(addr, list):
            ip_list = set(addr)
        elif isinstance(addr, str) and "," in addr:
            ip_list = {ip.strip() for ip in addr.split(",")}
        else:
            ip_list = {addr}
        ip_list = ip_list - all_ips  # remove IP addresses we've seen
        ip_entities.extend(IpAddress(Address=ip) for ip in ip_list)
        all_ips |= ip_list
        if geo_lookup:
            for ip_ent in ip_entities:
                geo_lite_lookup.lookup_ip(ip_entity=ip_ent)
    return ip_entities


@deprecated("Will be removed in a version 2.2", version="1.4.0")
@export
def create_ip_record(
    heartbeat_df: pd.DataFrame, az_net_df: pd.DataFrame = None
) -> IpAddress:
    """
    Generate ip_entity record for provided IP value.

    Parameters
    ----------
    heartbeat_df : pd.DataFrame
        A dataframe of heartbeat data for the host
    az_net_df : pd.DataFrame
        Option dataframe of Azure network data for the host

    Returns
    -------
    IP
        Details of the IP data collected

    """
    ip_entity = IpAddress()

    # Produce ip_entity record using available dataframes
    ip_hb = heartbeat_df.iloc[0]
    ip_entity.Address = ip_hb["ComputerIP"]
    ip_entity.hostname = ip_hb["Computer"]  # type: ignore
    ip_entity.SourceComputerId = ip_hb["SourceComputerId"]  # type: ignore
    ip_entity.OSType = ip_hb["OSType"]  # type: ignore
    ip_entity.OSName = ip_hb["OSName"]  # type: ignore
    ip_entity.OSVMajorVersion = ip_hb["OSMajorVersion"]  # type: ignore
    ip_entity.OSVMinorVersion = ip_hb["OSMinorVersion"]  # type: ignore
    ip_entity.ComputerEnvironment = ip_hb["ComputerEnvironment"]  # type: ignore
    ip_entity.OmsSolutions = [  # type: ignore
        sol.strip() for sol in ip_hb["Solutions"].split(",")
    ]
    ip_entity.VMUUID = ip_hb["VMUUID"]  # type: ignore
    ip_entity.SubscriptionId = ip_hb["SubscriptionId"]  # type: ignore
    geoloc_entity = GeoLocation()  # type: ignore
    geoloc_entity.CountryName = ip_hb["RemoteIPCountry"]  # type: ignore
    geoloc_entity.Longitude = ip_hb["RemoteIPLongitude"]  # type: ignore
    geoloc_entity.Latitude = ip_hb["RemoteIPLatitude"]  # type: ignore
    ip_entity.Location = geoloc_entity  # type: ignore

    # If Azure network data present add this to host record
    if az_net_df is not None and not az_net_df.empty:
        if len(az_net_df) == 1:
            priv_addr_str = az_net_df["PrivateIPAddresses"].loc[0]
            ip_entity["private_ips"] = convert_to_ip_entities(priv_addr_str)
            pub_addr_str = az_net_df["PublicIPAddresses"].loc[0]
            ip_entity["public_ips"] = convert_to_ip_entities(pub_addr_str)
        else:
            if "private_ips" not in ip_entity:
                ip_entity["private_ips"] = []
            if "public_ips" not in ip_entity:
                ip_entity["public_ips"] = []

    return ip_entity


@export  # noqa: MC0001
# pylint: disable=too-many-return-statements, invalid-name
def get_ip_type(ip: str = None, ip_str: str = None) -> str:  # noqa: MC0001
    """
    Validate value is an IP address and determine IPType category.

    (IPAddress category is e.g. Private/Public/Multicast).

    Parameters
    ----------
    ip : str
        The string of the IP Address
    ip_str : str
        The string of the IP Address - alias for `ip`

    Returns
    -------
    str
        Returns ip type string using ip address module

    """
    ip_str = ip or ip_str
    if not ip_str:
        raise ValueError("'ip' or 'ip_str' value must be specified")
    try:
        ipaddress.ip_address(ip_str)
    except ValueError:
        print(f"{ip_str} does not appear to be an IPv4 or IPv6 address")
    else:
        if ipaddress.ip_address(ip_str).is_multicast:
            return "Multicast"
        if ipaddress.ip_address(ip_str).is_global:
            return "Public"
        if ipaddress.ip_address(ip_str).is_loopback:
            return "Loopback"
        if ipaddress.ip_address(ip_str).is_link_local:
            return "Link Local"
        if ipaddress.ip_address(ip_str).is_unspecified:
            return "Unspecified"
        if ipaddress.ip_address(ip_str).is_private:
            return "Private"
        if ipaddress.ip_address(ip_str).is_reserved:
            return "Reserved"

    return "Unspecified"


# pylint: enable=too-many-return-statements


# pylint: disable=invalid-name
@deprecated("Replaced with ip_whois function", version="2.1.0")
@export
@lru_cache(maxsize=1024)
def get_whois_info(
    ip: str = None, show_progress: bool = False, **kwargs
) -> Tuple[str, dict]:
    """
    Retrieve whois ASN information for given IP address using IPWhois python package.

    Parameters
    ----------
    ip : str
        IP Address to look up.
    ip_str : str
        alias for `ip`.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    IP
        Details of the IP data collected

    Notes
    -----
    This function uses the Python functools lru_cache and
    will return answers from the cache for previously queried
    IP addresses.

    """
    ip_str = ip or kwargs.get("ip_str")
    if not ip_str:
        raise ValueError("'ip' or 'ip_str' value must be specified")
    ip_type = get_ip_type(ip_str)
    if ip_type == "Public":
        try:
            print(ip_str)
            whois_result = ip_whois(ip_str)
            if show_progress:
                print(".", end="")
            return whois_result  # type: ignore
        except MsticpyException as err:
            return f"Error during lookup of {ip_str} {type(err)}", {}
    return f"No ASN Information for IP type: {ip_type}", {}


# pylint: enable=invalid-name


@export
def get_whois_df(
    data: pd.DataFrame,
    ip_column: str,
    all_columns: bool = True,
    asn_col: str = "AsnDescription",
    whois_col: Optional[str] = "WhoIsData",
    show_progress: bool = False,
) -> pd.DataFrame:
    """
    Retrieve Whois ASN information for DataFrame of IP Addresses.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame
    ip_column : str
        Column name of IP Address to look up.
    all_columns:
        Expand all whois data to columns.
    asn_col : str, optional
        Name of the output column for ASN description,
        by default "ASNDescription".
        Ignored if `all_columns` is True.
    whois_col : str, optional
        Name of the output column for full whois data,
        by default "WhoIsData"
        Ignored if `all_columns` is True.
    show_progress : bool, optional
        Show progress for each query, by default False

    Returns
    -------
    pd.DataFrame
        Output DataFrame with results in added columns.

    """
    del show_progress
    whois_data = ip_whois(data[ip_column].drop_duplicates())
    if (
        isinstance(whois_data, pd.DataFrame)
        and not whois_data.empty
        and "query" in whois_data.columns
    ):
        data = data.merge(
            whois_data,  # type: ignore
            how="left",
            left_on=ip_column,
            right_on="query",
            suffixes=("", "_whois"),
        )
        data[whois_col] = data[whois_data.columns].apply(lambda x: x.to_dict(), axis=1)
        data[asn_col] = data["asn_description"]
        if not all_columns:
            return data.drop(columns=whois_data.columns)
        return data
    return data.assign(ASNDescription="No data returned")


@pd.api.extensions.register_dataframe_accessor("mp_whois")
@export
class IpWhoisAccessor:
    """Pandas api extension for IP Whois lookup."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def lookup(self, ip_column, **kwargs):
        """
        Extract IoCs from either a pandas DataFrame.

        Parameters
        ----------
        ip_column : str
            Column name of IP Address to look up.

        Other Parameters
        ----------------
        asn_col : str, optional
            Name of the output column for ASN description,
            by default "ASNDescription"
        whois_col : str, optional
            Name of the output column for full whois data,
            by default "WhoIsData"
        show_progress : bool, optional
            Show progress for each query, by default False

        Returns
        -------
        pd.DataFrame
            Output DataFrame with results in added columns.

        """
        warn_message = (
            "This accessor method has been deprecated.\n"
            "Please use IpAddress.util.whois() pivot function."
            "This will be removed in MSTICPy v2.2.0"
        )
        warnings.warn(warn_message, category=DeprecationWarning)
        return get_whois_df(data=self._df, ip_column=ip_column, **kwargs)


# pylint: disable=inconsistent-return-statements, invalid-name
def ip_whois(
    ip: Union[IpAddress, str, List, pd.Series, None] = None,
    ip_address: Union[IpAddress, str, List, pd.Series, None] = None,
    raw=False,
    query_rate: float = 0.5,
    retry_count: int = 5,
) -> Union[pd.DataFrame, Tuple]:
    # sourcery skip: assign-if-exp, reintroduce-else
    """
    Lookup IP Whois information.

    Parameters
    ----------
    ip : Union[IpAddress, str, List]
        An IP address or list of IP addresses to lookup.
    ip_address : Union[IpAddress, str, List]
        An IP address or list of IP addresses to lookup. Alias of `ip`.
    raw : bool, optional
        Set True if raw whois result wanted, by default False
    query_rate : float, optional
        Controls the rate at which queries are made, by default 0.5
    retry_count : int, optional
        The number of times to retry a query if it fails, default is 5

    Returns
    -------
    Union[pd.DataFrame, Dict]
        If a single IP Address provided result is a dictionary
        if multiple provided results are formatted as a dataframe.

    Raises
    ------
    ValueError:
        If neither `ip` nor `ip_address` is supplied.

    """
    ip = ip if ip is not None else ip_address
    if ip is None:
        raise ValueError("One of ip or ip_address parameters must be supplied.")
    if isinstance(ip, (list, pd.Series)):
        rate_limit = len(ip) > 50
        if rate_limit:
            print("Large number of lookups, this may take some time.")
        whois_results: Dict[str, Any] = {}
        for ip_addr in ip:
            if rate_limit:
                sleep(query_rate)
            whois_results[ip_addr] = _whois_lookup(  # type: ignore
                ip_addr, raw=raw, retry_count=retry_count
            ).properties
        return _whois_result_to_pandas(whois_results)
    if isinstance(ip, (str, IpAddress)):
        return _whois_lookup(ip, raw=raw)
    return pd.DataFrame()


def get_asn_details(asns: Union[str, List]) -> Union[pd.DataFrame, Dict]:
    """
    Get details about an ASN(s) from its number.

    Parameters
    ----------
    asns : Union[str, List]
        A single ASN or list of ASNs to lookup.

    Returns
    -------
    Union[pd.DataFrame, Dict]
        Details about the ASN(s) if a single ASN provided, result is a dictionary.
        If multiple provided, results are returned as a DataFrame.

    """
    if isinstance(asns, list):
        asn_detail_results = [_asn_results(str(asn)) for asn in asns]
        return pd.DataFrame(asn_detail_results)
    return _asn_results(str(asns))


def get_asn_from_name(name: str) -> Dict:
    """
    Get a list of ASNs that match a name.

    Parameters
    ----------
    name : str
        The name of the ASN to search for

    Returns
    -------
    Dict
        A list of ASNs that match the name.

    Raises
    ------
    MsticpyConnectionError
        If the list of all ASNs cannot be retrieved.
    MsticpyException
        If no matches found.

    """
    name = name.casefold()
    asns_dict = _ASNS_DICT()
    matches = {
        key: value for key, value in asns_dict.items() if name in value.casefold()
    }
    if len(matches.keys()) == 1:
        return next(iter(matches))  # type: ignore
    if len(matches.keys()) > 1:
        return matches

    raise MsticpyException(f"No ASNs found matching {name}")


def get_asn_from_ip(
    ip: Union[str, IpAddress, None] = None,
    ip_address: Union[str, IpAddress, None] = None,
) -> Dict:
    """
    Get the ASN that an IP belongs to.

    Parameters
    ----------
    ip : Union[str, IpAddress]
        IP address to lookup.
    ip_address : Union[str, IpAddress]
        IP address to lookup (alias of `ip`).

    Returns
    -------
    dict
        Details of the ASN that the IP belongs to.

    """
    ip = ip or ip_address
    if not ip:
        return {}
    if isinstance(ip, IpAddress):
        ip = ip.Address
    ip = ip.strip()
    query = f" -v {ip}\r\n"
    ip_response = _cymru_query(query)
    keys = ip_response.split("\n", maxsplit=1)[0].split("|")
    values = ip_response.split("\n")[1].split("|")
    return {key.strip(): value.strip() for key, value in zip(keys, values)}


class _IpWhoIsResult(NamedTuple):
    """Named tuple for IPWhoIs Result."""

    name: Optional[str]
    properties: Dict[str, Any] = {}


@lru_cache(maxsize=1024)
def _whois_lookup(
    ip_addr: Union[str, IpAddress], raw: bool = False, retry_count: int = 5  # type: ignore
) -> _IpWhoIsResult:
    """Conduct lookup of IP Whois information."""
    if isinstance(ip_addr, IpAddress):  # type: ignore
        ip_addr = ip_addr.Address
    asn_items = get_asn_from_ip(ip_addr.strip())
    registry_url: Optional[str] = None
    if asn_items and "Error: no ASN or IP match on line 1." not in asn_items:
        ipwhois_result = _IpWhoIsResult(asn_items["AS Name"], {})  # type: ignore
        ipwhois_result.properties["asn"] = asn_items["AS"]
        ipwhois_result.properties["query"] = asn_items["IP"]
        ipwhois_result.properties["asn_cidr"] = asn_items["BGP Prefix"]
        ipwhois_result.properties["asn_country_code"] = asn_items["CC"]
        ipwhois_result.properties["asn_registry"] = asn_items["Registry"]
        ipwhois_result.properties["asn_date"] = asn_items["Allocated"]
        ipwhois_result.properties["asn_description"] = asn_items["AS Name"]
        registry_url = _REGISTRIES.get(asn_items.get("Registry", ""), {}).get("url")
    if not asn_items or not registry_url:
        return _IpWhoIsResult(None)
    return _add_rdap_data(
        ipwhois_result=ipwhois_result,  # type: ignore
        rdap_reg_url=f"{registry_url}{ip_addr}",
        retry_count=retry_count,
        raw=raw,
    )


def _add_rdap_data(
    ipwhois_result: _IpWhoIsResult, rdap_reg_url: str, retry_count: int, raw: bool
) -> _IpWhoIsResult:
    """Add RDAP data to WhoIs result."""
    retries = 0
    while retries < retry_count:
        rdap_data = _rdap_lookup(url=rdap_reg_url, retry_count=retry_count)
        if rdap_data.status_code == 200:
            rdap_data_content = rdap_data.json()
            net = _create_net(rdap_data_content)
            ipwhois_result.properties["nets"] = [net]
            for link in rdap_data_content["links"]:
                if link["rel"] == "up":
                    up_data_link = link["href"]
                    up_rdap_data = httpx.get(up_data_link)
                    up_rdap_data_content = up_rdap_data.json()
                    up_net = _create_net(up_rdap_data_content)
                    ipwhois_result.properties["nets"].append(up_net)
            if raw:
                ipwhois_result.properties["raw"] = rdap_data
            break
        if rdap_data.status_code == 429:
            sleep(3)
            retries += 1
            continue
        raise MsticpyConnectionError(f"Error code: {rdap_data.status_code}")
    return ipwhois_result


def _rdap_lookup(url: str, retry_count: int = 5) -> httpx.Response:
    """Perform RDAP lookup with retries."""
    rdap_data = None
    while retry_count > 0 and not rdap_data:
        try:
            rdap_data = httpx.get(url)
        except (httpx.WriteError, httpx.ReadError):
            retry_count -= 1
    if not rdap_data:
        raise MsticpyException(
            "Rate limit exceeded - try adjusting query_rate parameter to slow down requests"
        )
    return rdap_data


def _whois_result_to_pandas(results: Union[str, List, Dict]) -> pd.DataFrame:
    """Transform whois results to a Pandas DataFrame."""
    if isinstance(results, dict):
        return pd.DataFrame(
            [result or {"query": idx} for idx, result in results.items()]
        )
    raise NotImplementedError("Only dict type current supported for `results`.")


def _find_address(
    entity: dict,
) -> Union[str, None]:  # pylint: disable=inconsistent-return-statements
    """Find an orgs address from an RDAP entity."""
    if "vcardArray" not in entity:
        return None
    for vcard in [vcard for vcard in entity["vcardArray"] if isinstance(vcard, list)]:
        for vcard_sub in vcard:
            if len(vcard) >= 2 and vcard_sub[0] == "adr" and "label" in vcard_sub[1]:
                return vcard_sub[1]["label"]
    return None


def _create_net(data: Dict) -> Dict:
    """Create a network object from RDAP data."""
    net_data = data.get("cidr0_cidrs", [None])[0] or {}
    net_prefixes = net_data.keys() & {"v4prefix", "v6prefix"}

    if not net_data or not net_prefixes:
        net_cidr = "No network data retrieved."
    else:
        net_cidr = " ".join(
            f"{net_data[net_prefix]}/{net_data.get('length', '')}"
            for net_prefix in net_prefixes
        )
    address = ""
    created = updated = None
    for item in data["events"]:
        created = item["eventDate"] if item["eventAction"] == "last changed" else None
        updated = item["eventDate"] if item["eventAction"] == "registration" else None
    for entity in data["entities"]:
        address = _find_address(entity)  # type: ignore
    regex = r"[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*"
    emails = re.findall(regex, str(data))
    return {
        "cidr": net_cidr,
        "handle": data["handle"],
        "name": data["name"],
        "startAddress": data["startAddress"],
        "endAddress": data["endAddress"],
        "created": created,
        "updated": updated,
        "address": address,
        "emails": emails,
    }


def _asn_whois_query(query, server, port=43, retry_count=5) -> str:
    """Connect to whois server and send query."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as conn:
        conn.connect((server, port))
        conn.send(query.encode())
        response = []
        response_data = None
        while retry_count > 0 and not response_data:
            try:
                response_data = conn.recv(4096).decode()
                if "error" in response_data:
                    raise MsticpyConnectionError(
                        "An error occurred during lookup, please try again."
                    )
                if "rate limit exceeded" in response_data:
                    raise MsticpyConnectionError(
                        "Rate limit exceeded please wait and try again."
                    )
                response.append(response_data)
            except (UnicodeDecodeError, ConnectionResetError):
                retry_count -= 1
                response_data = None
    return "".join(response)


def _cymru_query(query):
    """Query Cymru for ASN information."""
    return _asn_whois_query(query, "whois.cymru.com")


def _radb_query(query):
    """Query RADB for ASN information."""
    return _asn_whois_query(query, "whois.radb.net")


def _parse_asn_response(response) -> dict:
    """Parse ASN response into a dictionary."""
    response_output = {}
    for item in response.split("\n"):
        try:
            key = item.split(":")[0].strip()
            if key and key not in response_output:
                try:
                    response_output[key] = item.split(":")[1].strip()
                except IndexError:
                    continue
            elif key in response_output:
                if not isinstance(response_output[key], list):
                    response_output[key] = [response_output[key]]
                response_output[key].append(item.split(":")[1].strip())
        except IndexError:
            continue
    return response_output


def _asn_results(asn: str) -> dict:
    """Get ASN details from ASN number."""
    if not asn.startswith("AS"):
        asn = f"AS{asn}"
    query1 = f" {asn}\r\n"
    asn_response = _radb_query(query1)
    asn_details = _parse_asn_details(asn_response)
    query2 = f" -i origin {asn}\r\n"
    asn_ranges_response = _radb_query(query2)
    asn_details["ranges"] = _parse_asn_ranges(asn_ranges_response)
    return asn_details


def _parse_asn_details(response):
    """Parse ASN details response into a dictionary."""
    asn_keys = ["aut-num", "as-name", "descr", "notify", "changed"]
    asn_output = _parse_asn_response(response)
    asn_output_filtered = {
        key: value for key, value in asn_output.items() if key in asn_keys
    }
    asn_output_filtered["Autonomous Number"] = asn_output_filtered.pop("aut-num", None)
    asn_output_filtered["AS Name"] = asn_output_filtered.pop("as-name", None)
    asn_output_filtered["Description"] = asn_output_filtered.pop("descr", None)
    asn_output_filtered["Contact"] = asn_output_filtered.pop("notify", None)
    asn_output_filtered["Last Updated"] = asn_output_filtered.pop("changed", None)
    return asn_output_filtered


def _parse_asn_ranges(response):
    """Parse ASN ranges response into a list."""
    return [
        item.split(":   ")[1].strip()
        for item in response.split("\n")
        if item.split(":   ")[0].strip() == "route"
    ]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Generic Module for Lookup classes.

Input can be a single item or a pandas DataFrame containing
multiple items. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
import asyncio
import importlib
import warnings
from collections import ChainMap
from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, Union, cast

import nest_asyncio
import pandas as pd
from tqdm.auto import tqdm

from .._version import VERSION
from ..common.exceptions import MsticpyConfigError, MsticpyUserConfigError
from ..common.provider_settings import get_provider_settings, reload_settings
from ..common.utility import export, is_ipython
from ..vis.ti_browser import browse_results
from .lookup_result import LookupStatus

# used in dynamic instantiation of providers
from .provider_base import Provider, _make_sync

__version__ = VERSION
__author__ = "Florian Bracq"


class ProgressCounter:
    """Progress counter for async tasks."""

    def __init__(self, total: int):
        """Initialize the class."""
        self.total = total
        self._lock: asyncio.Condition = asyncio.Condition()
        self._remaining: int = total

    async def decrement(self, increment: int = 1):
        """Decrement the counter."""
        if self._remaining == 0:
            return
        async with self._lock:
            self._remaining -= increment

    async def get_remaining(self) -> int:
        """Get the current remaining count."""
        async with self._lock:
            return self._remaining


@export
class Lookup:
    """Item lookup from providers."""

    _NO_PROVIDERS_MSG = """
    No Providers are loaded - please check that
    you have correctly configured your msticpyconfig.yaml settings.
    """

    _HELP_URI = "https://msticpy.readthedocs.io/en/latest/DataEnrichment.html"

    PROVIDERS: Dict[str, Tuple[str, str]] = {}
    CUSTOM_PROVIDERS: Dict[str, Provider]

    PACKAGE: str = ""

    def __init__(self, providers: Optional[List[str]] = None, **kwargs):
        """
        Initialize TILookup instance.

        Parameters
        ----------
        primary_providers : Optional[List[Provider]], optional
            Primary Providers, by default None
        secondary_providers : Optional[List[Provider]], optional
            Secondary Providers, by default None
        providers: Optional[List[str]], optional
            List of provider names to load, by default all available
            providers are loaded. To see the list of available providers
            call `TILookup.list_available_providers()`.
            Note: if primary_provides or secondary_providers is specified
            This will override the providers list.

        """
        self._providers: Dict[str, Provider] = {}
        self._secondary_providers: Dict[str, Provider] = {}
        self._providers_to_load = providers

        primary_providers = kwargs.pop("primary_providers", None)
        if primary_providers:
            for prov in primary_providers:
                self.add_provider(prov, primary=True)
        secondary_providers = kwargs.pop("secondary_providers", None)
        if secondary_providers:
            warnings.warn(
                "'secondary_providers' is a deprecated parameter", DeprecationWarning
            )
            for prov in secondary_providers:
                self.add_provider(prov, primary=False)
        if not (primary_providers or secondary_providers):
            self._load_providers()

        self._all_providers = ChainMap(self._secondary_providers, self._providers)
        if is_ipython():
            nest_asyncio.apply()

    @property
    def loaded_providers(self) -> Dict[str, Provider]:
        """
        Return dictionary of loaded providers.

        Returns
        -------
        Dict[str, TIProvider]
            [description]

        """
        return dict(self._all_providers)

    @property
    def provider_status(self) -> Iterable[str]:
        """
        Return loaded provider status.

        Returns
        -------
        Iterable[str]
            List of providers and descriptions.

        """
        prim = [
            f"{prov_name} - {prov.description} (primary)"
            for prov_name, prov in self._providers.items()
        ]
        sec = [
            f"{prov_name} - {prov.description} (secondary)"
            for prov_name, prov in self._secondary_providers.items()
        ]
        return prim + sec

    @property
    def configured_providers(self) -> List[str]:
        """
        Return a list of available providers that have configuration details present.

        Returns
        -------
        List[str]
            List of TI Provider classes.

        """
        prim_conf = list(self._providers.keys())
        sec_conf = list(self._secondary_providers.keys())

        return prim_conf + sec_conf

    def enable_provider(self, providers: Union[str, Iterable[str]]):
        """
        Set the provider(s) as primary (used by default).

        Parameters
        ----------
        providers : Union[str, Iterable[str]
            Provider name or list of names.
            Use `list_available_providers()` to see the list of loaded providers.

        Raises
        ------
        ValueError
            If the provider name is not recognized.

        """
        provs_to_enable = [providers] if isinstance(providers, str) else providers
        for provider in provs_to_enable:
            if provider in self._secondary_providers:
                self._providers[provider] = self._secondary_providers[provider]
                del self._secondary_providers[provider]
            elif provider not in self._providers:
                raise ValueError(
                    f"Unknown provider '{provider}'. Available providers:",
                    ", ".join(self.list_available_providers(as_list=True)),  # type: ignore
                )

    def disable_provider(self, providers: Union[str, Iterable[str]]):
        """
        Set the provider as secondary (not used by default).

        Parameters
        ----------
        providers : Union[str, Iterable[str]
            Provider name or list of names.
            Use `list_available_providers()` to see the list of loaded providers.

        Raises
        ------
        ValueError
            If the provider name is not recognized.

        """
        provs_to_disable = [providers] if isinstance(providers, str) else providers
        for provider in provs_to_disable:
            if provider in self._providers:
                self._secondary_providers[provider] = self._providers[provider]
                del self._providers[provider]
            elif provider not in self._secondary_providers:
                raise ValueError(
                    f"Unknown provider '{provider}'. Available providers:",
                    ", ".join(self.list_available_providers(as_list=True)),  # type: ignore
                )

    def set_provider_state(self, prov_dict: Dict[str, bool]):
        """
        Set a dict of providers to primary/secondary.

        Parameters
        ----------
        prov_dict : Dict[str, bool]
            Dictionary of provider name and bool - True if enabled/primary,
            False if disabled/secondary.

        """
        for prov, state in prov_dict.items():
            if state:
                self.enable_provider(prov)
            else:
                self.disable_provider(prov)

    @classmethod
    def browse_results(
        cls, data: pd.DataFrame, severities: Optional[List[str]] = None, **kwargs
    ):
        """
        Return TI Results list browser.

        Parameters
        ----------
        data : pd.DataFrame
            TI Results data from TIProviders
        severities : Optional[List[str]], optional
            A list of the severity classes to show.
            By default these are ['warning', 'high'].
            Pass ['information', 'warning', 'high'] to see all
            results.

        Other Parameters
        ----------------
        kwargs :
            passed to SelectItem constructor.

        Returns
        -------
        SelectItem
            SelectItem browser for TI Data.

        """
        if not isinstance(data, pd.DataFrame):
            print("Input data is in an unexpected format.")
            return None
        return browse_results(data=data, severities=severities, **kwargs)

    browse = browse_results

    def provider_usage(self):
        """Print usage of loaded providers."""
        print("Primary providers")
        print("-----------------")
        if self._providers:
            for prov_name, prov in self._providers.items():
                print(f"\nProvider class: {prov_name}")
                prov.usage()
        else:
            print("none")
        print("\nSecondary providers")
        print("-------------------")
        if self._secondary_providers:
            for prov_name, prov in self._secondary_providers.items():
                print(f"\nProvider class: {prov_name}")
                prov.usage()
        else:
            print("none")

    @classmethod
    def reload_provider_settings(cls):
        """Reload provider settings from config."""
        reload_settings()
        print(
            "Settings reloaded. Use reload_providers to update settings",
            "for loaded providers.",
        )

    def reload_providers(self):
        """Reload settings and provider classes."""
        reload_settings()
        self._load_providers()

    def add_provider(self, provider: Provider, name: str = None, primary: bool = True):
        """
        Add a provider to the current collection.

        Parameters
        ----------
        provider : TIProvider
            Provider instance
        name : str, optional
            The name to use for the provider (overrides the class name
            of `provider`)
        primary : bool, optional
            "primary" or "secondary" if False, by default "primary"

        """
        if not name:
            name = provider.__class__.__name__
        if primary:
            self._providers[name] = provider
        else:
            self._secondary_providers[name] = provider

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-arguments
    def lookup_item(
        self,
        item: str,
        item_type: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup single item in active providers.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            One of ItemExtract.ItemType, by default None
            If none, the Item type will be inferred
        query_type: str, optional
            The query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return self.lookup_items(
            data={item: item_type},
            query_type=query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def lookup_items(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        item_col: str = None,
        item_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Mapping (e.g. a dict) of [item, ItemType]
            3. Iterable of items - ItemTypes will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
            ("col" and "column" are also aliases for this parameter)
        item_type_col : str, optional
            DataFrame column to use for ItemTypes, by default None
        query_type: str, optional
            The item query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"

        Other Parameters
        ----------------
        progress : bool
            Use progress bar to track completion, by default True
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return _make_sync(
            self._lookup_items_async(
                data=data,
                item_col=item_col,
                item_type_col=item_type_col,
                query_type=query_type,
                providers=providers,
                default_providers=default_providers,
                prov_scope=prov_scope,
                **kwargs,
            )
        )

    @staticmethod
    def result_to_df(item_lookup: pd.DataFrame) -> pd.DataFrame:
        """
        Return DataFrame representation of Lookup response.

        Parameters
        ----------
        item_lookup : pd.DataFrame
            Output from `lookup_item`

        Returns
        -------
        pd.DataFrame
            The response as a DataFrame with a row for each
            provider response.

        """
        if isinstance(item_lookup, pd.DataFrame):
            return item_lookup
        raise ValueError(f"DataFrame was expected, but {type(item_lookup)} received.")

    # pylint: disable=too-many-locals
    async def _lookup_items_async(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        item_col: str = None,
        item_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """Lookup items async."""
        item_col = item_col or kwargs.pop("col", kwargs.pop("column", None))
        progress = kwargs.pop("progress", True)
        selected_providers = self._select_providers(
            providers or default_providers, prov_scope
        )
        if not selected_providers:
            raise MsticpyUserConfigError(
                self._NO_PROVIDERS_MSG,
                title="No Threat Intel Provider configuration found.",
                help_uri=self._HELP_URI,
            )

        event_loop = asyncio.get_event_loop()
        result_futures: List[Any] = []
        provider_names: List[str] = []

        prog_counter = ProgressCounter(
            total=len(data) * len(selected_providers)  # type: ignore
        )

        # create a list of futures/tasks to await
        for prov_name, provider in selected_providers.items():
            provider_names.append(prov_name)
            result_futures.append(
                provider.lookup_items_async(
                    data=data,
                    item_col=item_col,
                    item_type_col=item_type_col,
                    query_type=query_type,
                    prog_counter=prog_counter if progress else None,
                    **kwargs,
                )
            )
        if progress:
            # Create a task for tqdm
            prog_task = event_loop.create_task(self._track_completion(prog_counter))
        # collect the return values of the tasks
        results = await asyncio.gather(*result_futures)
        # cancel the progress task if results have completed.
        if progress:
            prog_task.cancel()
        return self._combine_results(results, provider_names, **kwargs)

    def lookup_items_sync(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        item_col: str = None,
        item_type_col: str = None,
        query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of items.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `item_col` parameter)
            2. Mapping (e.g. a dict) of [item, ItemType]
            3. Iterable of items - ItemTypes will be inferred
        item_col : str, optional
            DataFrame column to use for items, by default None
            ("col" and "column" are also aliases for this parameter)
        item_type_col : str, optional
            DataFrame column to use for ItemTypes, by default None
        query_type: str, optional
            The item query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        item_col = item_col or kwargs.pop("col", kwargs.pop("column", None))

        selected_providers = self._select_providers(
            providers or default_providers, prov_scope
        )
        if not selected_providers:
            raise MsticpyUserConfigError(
                self._NO_PROVIDERS_MSG,
                title="No Threat Intel Provider configuration found.",
                help_uri=self._HELP_URI,
            )

        results: List[Any] = []
        provider_names: List[str] = []
        for prov_name, provider in selected_providers.items():
            provider_names.append(prov_name)
            results.append(
                provider.lookup_items(
                    data=data,
                    item_col=item_col,
                    item_type_col=item_type_col,
                    query_type=query_type,
                    **kwargs,
                )
            )
        return self._combine_results(results, provider_names, **kwargs)

    @staticmethod
    async def _track_completion(prog_counter):
        total = await prog_counter.get_remaining()
        with tqdm(total=total, unit="obs", desc="Observables processed") as prog_bar:
            try:
                last_remaining = total
                while last_remaining:
                    new_remaining = await prog_counter.get_remaining()
                    incr = last_remaining - new_remaining
                    if incr:
                        prog_bar.update(incr)
                    last_remaining = new_remaining
                    # print(f"progress: incr {incr}, last: {last_remaining}")
                    await asyncio.sleep(0)
            except asyncio.CancelledError:
                # make progress bar get to 100% on cancel
                final_remaining = await prog_counter.get_remaining()
                if final_remaining:
                    prog_bar.update(total - final_remaining)

    @property
    def available_providers(self) -> List[str]:
        """
        Return a list of builtin and plugin providers.

        Returns
        -------
        List[str]
            List of TI Provider classes.

        """
        return list(self.PROVIDERS) + list(self.CUSTOM_PROVIDERS)

    @classmethod
    def list_available_providers(
        cls, show_query_types=False, as_list: bool = False
    ) -> Optional[List[str]]:  # type: ignore
        """
        Print a list of builtin providers with optional usage.

        Parameters
        ----------
        show_query_types : bool, optional
            Show query types supported by providers, by default False
        as_list : bool, optional
            Return list of providers instead of printing to stdout.
            Note: if you specify `show_query_types` this will be printed
            irrespective of this parameter setting.

        Returns
        -------
        Optional[List[str]]
            A list of provider names (if `return_list=True`)

        """
        providers = []
        for provider_name in cls.PROVIDERS:
            provider_class = cls.import_provider(provider_name)
            if not as_list:
                print(provider_name)
            providers.append(provider_name)
            if show_query_types and provider_class:
                provider_class.usage()

        return providers if as_list else None

    @classmethod
    def import_provider(cls, provider: str) -> Provider:
        """Import provider class."""
        mod_name, cls_name = cls.PROVIDERS.get(provider, (None, None))

        if not (mod_name and cls_name):
            if hasattr(cls, "CUSTOM_PROVIDERS") and provider in cls.CUSTOM_PROVIDERS:
                return cast(Provider, cls.CUSTOM_PROVIDERS[provider])
            raise LookupError(
                f"No provider named '{provider}'.",
                "Possible values are:",
                ", ".join(list(cls.PROVIDERS) + list(cls.CUSTOM_PROVIDERS)),
            )

        imp_module = importlib.import_module(
            f"msticpy.context.{cls.PACKAGE}.{mod_name}", package="msticpy"
        )
        return getattr(imp_module, cls_name)

    def _load_providers(self, **kwargs):
        """Load provider classes based on config."""
        prov_type = kwargs.get("providers", "Providers")
        prov_settings = get_provider_settings(prov_type)
        provider_class: Provider
        for provider_name, settings in prov_settings.items():
            if (
                self._providers_to_load is not None
                and provider_name not in self._providers_to_load
            ) or settings.provider == "--no-load--":
                continue
            try:
                provider_name = settings.provider or provider_name
                provider_class = self.import_provider(provider_name)
            except LookupError:
                warnings.warn(
                    f"Could not find provider class for {provider_name} "
                    f"in config section '{provider_name}'. "
                    f"Provider class name in config is '{settings.provider}'"
                )
                continue

            # instantiate class sending args from settings to init
            try:
                provider_instance = provider_class(**(settings.args))
            except MsticpyConfigError as mp_ex:
                # If the TI Provider didn't load, raise an exception
                raise MsticpyUserConfigError(
                    f"Could not load Provider {provider_name}",
                    *mp_ex.args,
                    "To avoid loading this provider please use the 'providers' parameter"
                    + " to specify which providers to load.",
                    title="Provider configuration error",
                    help_uri=self._HELP_URI,
                ) from mp_ex

            # set the description from settings, if one is provided, otherwise
            # use class docstring.
            provider_instance.description = (
                settings.description or provider_instance.__doc__
            )

            self.add_provider(
                provider=provider_instance, name=provider_name, primary=settings.primary
            )

    def _select_providers(
        self, providers: List[str] = None, prov_scope: str = "primary"
    ) -> Dict[str, Provider]:
        """
        Return required subset of providers.

        Parameters
        ----------
        providers : List[str], optional
            Explicit list of provider names, by default None
        prov_scope : str, optional
            Provider scope, by default "primary"
            Other values are "all" and "secondary"

        Returns
        -------
        Dict[str, TIProvider]
            Dictionary of provider names and instances.

        """
        if providers:
            return {
                prov_name: prov
                for prov_name, prov in self._all_providers.items()
                if prov_name in providers
            }
        if prov_scope == "all":
            return dict(self._all_providers)
        if prov_scope == "primary":
            return self._providers
        return self._secondary_providers

    @staticmethod
    def _combine_results(
        results: Iterable[pd.DataFrame], provider_names: List[str], **kwargs
    ):
        """Combine dataframe results into single DF."""
        result_list: List[pd.DataFrame] = []
        for prov_name, provider_result in zip(provider_names, results):
            if provider_result is None or provider_result.empty:
                continue
            if not kwargs.get("show_not_supported", False):
                provider_result = provider_result[
                    provider_result["Status"] != LookupStatus.NOT_SUPPORTED.value
                ]
            if not kwargs.get("show_bad_item", False):
                provider_result = provider_result[
                    provider_result["Status"] != LookupStatus.BAD_FORMAT.value
                ]
            provider_result["Provider"] = prov_name
            result_list.append(provider_result)

        if not result_list:
            print("No Item matches")
        return pd.concat(result_list, sort=False) if result_list else None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for timeseries analysis functions."""
import inspect
from datetime import datetime
from typing import Dict, List, Optional

import pandas as pd

from .._version import VERSION
from ..common.exceptions import MsticpyException, MsticpyImportExtraError
from ..common.timespan import TimeSpan
from ..common.utility import check_kwargs, export
from ..vis.timeseries import display_timeseries_anomalies

try:
    from scipy import stats
    from statsmodels.tsa.seasonal import STL
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without scipy and statsmodel installed",
        title="Error importing package",
        extra="ml",
    ) from imp_err

__version__ = VERSION
__author__ = "Ashwin Patil"


@pd.api.extensions.register_dataframe_accessor("mp_timeseries")
class MsticpyTimeSeriesAccessor:
    """Msticpy pandas accessor for time series functions."""

    def __init__(self, pandas_obj):
        """Initialize the extension."""
        self._df = pandas_obj

    def analyze(self, **kwargs) -> pd.DataFrame:
        """
        Return anomalies in Timeseries using STL.

        Parameters
        ----------
        time_column : str, optional
            If the input data is not indexed on the time column, use this column
            as the time index
        data_column : str, optional
            Use named column if the input data has more than one column.
        seasonal : int, optional
            Seasonality period of the input data required for STL.
            Must be an odd integer, and should normally be >= 7 (default).
        period: int, optional
            Periodicity of the the input data. by default 24 (Hourly).
        score_threshold : float, optional
            standard deviation threshold value calculated using Z-score used to
            flag anomalies, by default 3

        Returns
        -------
        pd.DataFrame
            Returns a dataframe with additional columns by decomposing time series data
            into residual, trend, seasonal, weights, baseline, score and anomalies.
            The anomalies column will have 0, 1,-1 values based on score_threshold set.

        Notes
        -----
        The decomposition method is STL - Seasonal-Trend Decomposition using LOESS

        """
        return ts_anomalies_stl(self._df, **kwargs)

    def anomaly_periods(self, **kwargs):
        """
        Return list of anomaly period as TimeSpans.

        Parameters
        ----------
        time_column : str, optional
            The name of the time column
        period : str, optional
            pandas-compatible time period designator,
            by default "1H"
        pos_only : bool, optional
            If True only extract positive anomaly periods,
            else extract both positive and negative.
            By default, True
        anomalies_column : str, optional
            The column containing the anomalies flag.

        Returns
        -------
        List[TimeSpan] :
            TimeSpan(start, end)

        """
        return find_anomaly_periods(data=self._df, **kwargs)

    def apply_threshold(self, **kwargs):
        """
        Return DataFrame with anomalies calculated based on new threshold.

        Parameters
        ----------
        threshold : float
            Threshold above (beyond) which values will be marked as
            anomalies. Used as positive and negative threshold
            unless `threshold_low` is specified.
        threshold_low : Optional[float], optional
            The threshold below which values will be reported
            as anomalies, by default None.
        anomalies_column : str, optional
            The column containing the anomalies flag.

        Returns
        -------
        pd.DataFrame
            Output DataFrame with recalculated anomalies.

        """
        return set_new_anomaly_threshold(data=self._df, **kwargs)

    def kql_periods(self, **kwargs):
        """
        Return KQL filter expression for anomaly time periods.

        Parameters
        ----------
        time_column : str, optional
            The name of the time column
        period : str, optional
            pandas-compatible time period designator,
            by default "1H"
        pos_only : bool, optional
            If True only extract positive anomaly periods,
            else extract both positive and negative.
            By default, True

        Returns
        -------
        List[TimeSpan] :
            TimeSpan(start, end)

        """
        anom_periods = extract_anomaly_periods(data=self._df, **kwargs)
        return create_time_period_kqlfilter(anom_periods)

    def plot(self, **kwargs):
        """
        Display time series anomalies visualization.

        Parameters
        ----------
            value_column : str, optional
            Name of column holding numeric values to plot against time series to
            determine anomalies
            (the default is 'Total')
        y : str, optional
            alias for "value_column"
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        anomalies_column : str, optional
            Name of the column holding binary status(1/0) for anomaly/benign
            (the default is 'anomalies')
        period : int, optional
            Period of the dataset for hourly-no of days, for daily-no of weeks.
            This is used to correctly calculate the plot height.
            (the default is 30)

        Other Parameters
        ----------------
        ref_time : datetime, optional
            Input reference line to display (the default is None)
        title : str, optional
            Title to display (the default is None)
        legend: str, optional
            Where to position the legend
            None, left, right or inline (default is None)
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        xgrid : bool, optional
            Whether to show the xaxis grid (default is True)
        ygrid : bool, optional
            Whether to show the yaxis grid (default is False)
        color : list, optional
            List of colors to use in 3 plots as specified in order
            3 plots- line(observed), circle(baseline), circle_x/user specified(anomalies).
            (the default is ["navy", "green", "firebrick"])

        Returns
        -------
        figure
            The bokeh plot figure.

        """
        return display_timeseries_anomalies(data=self._df, **kwargs)


# Constants
_DEFAULT_KWARGS = [
    "seasonal",
    "period",
    "score_threshold",
    "time_column",
    "data_column",
]


@export
def ts_anomalies_stl(data: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Return anomalies in Timeseries using STL.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame as a time series data set retrieved from data connector or
        external data source. Dataframe must have 2 columns with time column
        set as index and other numeric value.

    Other Parameters
    ----------------
    time_column : str, optional
        If the input data is not indexed on the time column, use this column
        as the time index
    data_column : str, optional
        Use named column if the input data has more than one column.
    seasonal : int, optional
        Seasonality period of the input data required for STL.
        Must be an odd integer, and should normally be >= 7 (default).
    period: int, optional
        Periodicity of the the input data. by default 24 (Hourly).
    score_threshold : float, optional
        standard deviation threshold value calculated using Z-score used to
        flag anomalies, by default 3

    Returns
    -------
    pd.DataFrame
        Returns a dataframe with additional columns by decomposing time series data
        into residual, trend, seasonal, weights, baseline, score and anomalies.
        The anomalies column will have 0, 1,-1 values based on score_threshold set.

    Notes
    -----
    The decomposition method is STL - Seasonal-Trend Decomposition using LOESS

    """
    check_kwargs(kwargs, _DEFAULT_KWARGS)
    seasonal: int = kwargs.get("seasonal", 7)
    period: int = kwargs.get("period", 24)
    score_threshold: float = kwargs.get("score_threshold", 3.0)
    time_column = kwargs.get("time_column")
    data_column = kwargs.get("data_column")

    if not isinstance(data, pd.DataFrame):
        raise MsticpyException("input data should be a pandas dataframe")

    if time_column:
        data = data.set_index(time_column)
    data_column = data_column or data.columns[0]
    data = data[[data_column]]

    # STL method does Season-Trend decomposition using LOESS.
    # Accepts timeseries dataframe
    stl = STL(data[data_column].values, seasonal=seasonal, period=period)
    # Fitting the data - Estimate season, trend and residuals components.
    res = stl.fit()
    result = data.copy()
    # Create dataframe columns from decomposition results
    result["residual"] = res.resid
    result["trend"] = res.trend
    result["seasonal"] = res.seasonal
    result["weights"] = res.weights
    # Baseline is generally seasonal + trend
    result["baseline"] = result["seasonal"] + result["trend"]
    # Type cast and replace na values with 0
    result = result.fillna(0).astype("int64")
    # Calculate zscore based on residual column
    # this column does not contain seasonal/trend components
    result["score"] = stats.zscore(result["residual"])
    # create spikes(1) and dips(-1) based on threshold and seasonal columns
    result.loc[
        (result["score"] > score_threshold) & (result["seasonal"] > 0), "anomalies"
    ] = 1
    result.loc[
        (result["score"] > score_threshold) & (result["seasonal"] < 0), "anomalies"
    ] = -1
    result.loc[(result["score"] < score_threshold), "anomalies"] = 0
    # Datatype casting
    result["anomalies"] = result["anomalies"].astype("int64")

    time_index_name = data.index.name or "index"
    return result.reset_index().sort_values(time_index_name, ascending=True)


timeseries_anomalies_stl = ts_anomalies_stl


def extract_anomaly_periods(
    data: pd.DataFrame,
    time_column: str = "TimeGenerated",
    period: str = "1h",
    pos_only: bool = True,
    anomalies_column: str = "anomalies",
) -> Dict[datetime, datetime]:
    """
    Return dictionary of anomaly periods, merging adjacent ones.

    Parameters
    ----------
    data : pd.DataFrame
        The data to process
    time_column : str, optional
        The name of the time column
    period : str, optional
        pandas-compatible time period designator,
        by default "1h"
    pos_only : bool, optional
        If True only extract positive anomaly periods,
        else extract both positive and negative.
        By default, True
    anomalies_column : str, optional
        The column containing the anomalies flag.

    Returns
    -------
    Dict[datetime, datetime] :
        start_period, end_period

    """
    # Resample data based on period - period is the granularity that
    # we want to merge 2 adjacent samples on.
    anom_filter = [1] if pos_only else [1, -1]
    resampled = (
        data[(data[anomalies_column].isin(anom_filter))]
        .sort_values(time_column)
        .set_index(time_column)
        .resample(period)
    )

    end_period = None
    start_period = None
    periods = {}

    # iterate through the resampled data
    for time, group in resampled:
        if group.empty:
            continue
        if not end_period:
            # If we're not already in an anomaly period
            # create start/end for a new one
            start_period = time - pd.Timedelta(period)  # type: ignore
            end_period = time + pd.Timedelta(period)  # type: ignore
            periods[start_period] = end_period
        elif (time - end_period) <= pd.Timedelta(
            period
        ) * 2 and start_period is not None:
            # if the current time is less than 2x the period away
            # from our current end_period time, update the end_time
            periods[start_period] = time + pd.Timedelta(period)
        else:
            # otherwise start a new period
            start_period = time - pd.Timedelta(period)
            periods[start_period] = time + pd.Timedelta(period)
        end_period = time
    return periods


def find_anomaly_periods(
    data: pd.DataFrame,
    time_column: str = "TimeGenerated",
    period: str = "1h",
    pos_only: bool = True,
    anomalies_column: str = "anomalies",
) -> List[TimeSpan]:
    """
    Return list of anomaly period as TimeSpans.

    Parameters
    ----------
    data : pd.DataFrame
        The data to process
    time_column : str, optional
        The name of the time column
    period : str, optional
        pandas-compatible time period designator,
        by default "1h"
    pos_only : bool, optional
        If True only extract positive anomaly periods,
        else extract both positive and negative.
        By default, True
    anomalies_column : str, optional
        The column containing the anomalies flag.

    Returns
    -------
    List[TimeSpan] :
        TimeSpan(start, end)

    """
    return [
        TimeSpan(start=key, end=val)
        for key, val in extract_anomaly_periods(
            data=data,
            time_column=time_column,
            period=period,
            pos_only=pos_only,
            anomalies_column=anomalies_column,
        ).items()
    ]


def create_time_period_kqlfilter(periods: Dict[datetime, datetime]) -> str:
    """
    Return KQL time filter expression from anomaly periods.

    Parameters
    ----------
    periods : Dict[datetime, datetime]
        Dict of start, end periods

    Returns
    -------
    str
        KQL filter clause

    """
    time_column = "TimeGenerated"
    time_brackets = [
        f"{time_column} between (datetime({start}) .. datetime({end}))"
        for start, end in periods.items()
    ]

    return f"| where {' or '.join(time_brackets)}"


def set_new_anomaly_threshold(
    data: pd.DataFrame,
    threshold: float,
    threshold_low: Optional[float] = None,
    anomalies_column: str = "anomalies",
) -> pd.DataFrame:
    """
    Return DataFrame with anomalies calculated based on new threshold.

    Parameters
    ----------
    data : pd.DataFrame
        Input DataFrame
    threshold : float
        Threshold above (beyond) which values will be marked as
        anomalies. Used as positive and negative threshold
        unless `threshold_low` is specified.
    threshold_low : Optional[float], optional
        The threshhold below which values will be reported
        as anomalies, by default None.
    anomalies_column : str, optional
        The column containing the anomalies flag.

    Returns
    -------
    pd.DataFrame
        Output DataFrame with recalculated anomalies.

    """
    new_anoms_col = "__new_anomalies__"
    threshold_low = threshold_low or threshold
    new_df = data.assign(__new_anomalies__=0)
    new_df.loc[new_df["score"] >= threshold, new_anoms_col] = 1
    new_df.loc[new_df["score"] <= -threshold_low, new_anoms_col] = -1
    return new_df.drop(columns=[anomalies_column]).rename(
        columns={new_anoms_col: anomalies_column}
    )


def _doc_remove_data(func, remove_other_params: bool = False):
    new_doc = inspect.getdoc(func)
    if not new_doc:
        return ""
    doc_lines = iter(new_doc.split("\n"))
    out_lines = []
    while True:
        try:
            line = next(doc_lines)
            if line.startswith("data"):
                line = next(doc_lines)
                while line.startswith("    "):
                    line = next(doc_lines)
                out_lines.append(line)
                continue
            if remove_other_params and line.startswith("Other Parameters"):
                line = next(doc_lines)
                continue
            out_lines.append(line)
        except StopIteration:
            break
    return "\n".join(out_lines)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for TILookup classes.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""

from abc import abstractmethod
from typing import Any, Dict, Iterable, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.utility import export
from ..provider_base import PivotProvider, Provider
from .result_severity import ResultSeverity

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class TIProvider(Provider):
    """Abstract base class for Threat Intel providers."""

    _QUERIES: Dict[str, Any] = {}

    def _check_item_type(
        self, item: str, item_type: str = None, query_subtype: str = None
    ) -> Dict:
        """
        Check Item Type and cleans up item.

        Parameters
        ----------
        item : str
            item
        item_type : str, optional
            item type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed
            item.
            Status is none-zero on failure.

        """
        return self._check_ioc_type(
            item,
            item_type,
            query_subtype,
        )

    def _check_ioc_type(
        self, ioc: str, ioc_type: str = None, query_subtype: str = None
    ) -> Dict:
        """
        Check Ioc Type and cleans up ioc.

        Parameters
        ----------
        ioc : str
            IoC
        ioc_type : str, optional
            IoC type, by default None
        query_subtype : str, optional
            Query sub-type, if any, by default None

        Returns
        -------
        Dict
            Dict result with resolved type and pre-processed
            Ioc.
            Status is none-zero on failure.

        """
        result = super()._check_item_type(
            item=ioc, item_type=ioc_type, query_subtype=query_subtype
        )
        result["Ioc"] = result.pop("Item")
        result["IocType"] = result.pop("ItemType")
        result["SafeIoc"] = result.pop("SanitizedValue")
        result["Severity"] = ResultSeverity.information.name
        return result

    def lookup_item(
        self, item: str, item_type: str = None, query_type: str = None, **kwargs
    ) -> pd.DataFrame:
        """
        Lookup a single item.

        Parameters
        ----------
        item : str
            Item value to lookup
        item_type : str, optional
            The Type of the value to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular observable to try avoid repeated network calls for
        the same item.

        """
        return self.lookup_ioc(
            ioc=item,
            ioc_type=item_type,
            query_type=query_type,
            **kwargs,
        )

    @abstractmethod
    def parse_results(self, response: Dict) -> Tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """

    @abstractmethod
    def lookup_ioc(
        self,
        ioc: str,
        ioc_type: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """

    def lookup_iocs(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup collection of IoC observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `ioc_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return self.lookup_items(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
            **kwargs,
        )

    async def lookup_iocs_async(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Call base async wrapper."""
        return await self._lookup_items_async_wrapper(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
            **kwargs,
        )

    async def _lookup_iocs_async_wrapper(
        self,
        data: Union[pd.DataFrame, Dict[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        query_type: str = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Async wrapper for providers that do not implement lookup_iocs_async.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        return await self._lookup_items_async_wrapper(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=query_type,
            **kwargs,
        )

    @property
    def ioc_query_defs(self) -> Dict[str, Any]:
        """
        Return current dictionary of IoC query/request definitions.

        Returns
        -------
        Dict[str, Any]
            IoC query/request definitions keyed by IoCType

        """
        return self._QUERIES

    @classmethod
    def usage(cls):
        """Print usage of provider."""
        print(f"{cls.__doc__} Supported query types:")
        for ioc_key in sorted(cls._QUERIES):
            ioc_key_elems = ioc_key.split("-", maxsplit=1)
            if len(ioc_key_elems) == 1:
                print(f"\tioc_type={ioc_key_elems[0]}")
            if len(ioc_key_elems) == 2:
                print(f"\tioc_type={ioc_key_elems[0]}, query_type={ioc_key_elems[1]}")

    @staticmethod
    def resolve_ioc_type(observable: str) -> str:
        """
        Return IoCType determined by IoCExtract.

        Parameters
        ----------
        observable : str
            IoC observable string

        Returns
        -------
        str
            IoC Type (or unknown if type could not be determined)

        """
        return TIProvider.resolve_item_type(observable)


class TIPivotProvider(PivotProvider):
    """A class which provides TI pivot functions and a means of registering them."""

    @abstractmethod
    def register_pivots(
        self,
        pivot_reg: "PivotRegistration",  # type: ignore # noqa: F821
        pivot: "Pivot",  # type: ignore # noqa: F821
    ):
        """
        Register pivot functions for the TI Provider.

        Parameters
        ----------
        pivot_reg : PivotRegistration
            Pivot registration settings.
        pivot : Pivot
            Pivot library instance

        """



#  -------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for
#  license information.
#  --------------------------------------------------------------------------
"""Splunk Driver class."""
import logging
from datetime import datetime, timedelta, timezone
from time import sleep
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import jwt
import pandas as pd
from tqdm import tqdm

from ..._version import VERSION
from ...common.exceptions import (
    MsticpyConnectionError,
    MsticpyDataQueryError,
    MsticpyImportExtraError,
    MsticpyUserConfigError,
)
from ...common.utility import check_kwargs, export
from ..core.query_defns import Formatters
from .driver_base import DriverBase, DriverProps, QuerySource

try:
    import splunklib.client as sp_client
    import splunklib.results as sp_results
    from splunklib.client import AuthenticationError, HTTPError
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without splunk-sdk installed",
        title="Error importing splunk-sdk",
        extra="splunk",
    ) from imp_err

__version__ = VERSION
__author__ = "Ashwin Patil, Tatsuya Hasegawa"

logger = logging.getLogger(__name__)


SPLUNK_CONNECT_ARGS = {
    "host": "(string) The host name (the default is 'localhost').",
    "port": "(string) The port number (the default is '8089').",
    "http_scheme": (
        "('https' or 'http') The scheme for accessing the service "
        + "(the default is 'https')."
    ),
    "verify": (
        "(Boolean) Enable (True) or disable (False) SSL verrification for "
        + "https connections. (optional, the default is True)"
    ),
    "owner": "(string) The owner context of the namespace (optional).",
    "app": "(string) The app context of the namespace (optional).",
    "sharing": (
        "('global', 'system', 'app', or 'user') "
        + "The sharing mode for the namespace (the default is 'user')."
    ),
    "token": (
        "(string) The current session token (optional). Session tokens can be"
        + " shared across multiple service instances."
    ),
    "cookie": (
        "(string) A session cookie. When provided, you don’t need to call"
        + " login(). This parameter is only supported for Splunk 6.2+."
    ),
    "autologin": (
        "(boolean) When True, automatically tries to log in again if"
        + " the session terminates."
    ),
    "username": (
        "(string) The Splunk account username, which is used to "
        + "authenticate the Splunk instance."
    ),
    "password": "(string) The password for the Splunk account.",
    "bearer_token": (
        "(string) The Authorization Bearer Token <JWT> created in the Splunk."
        + " (Named as 'splunkToken' in Splunk Python SDK)"
    ),
}


@export
class SplunkDriver(DriverBase):
    """Driver to connect and query from Splunk."""

    _SPLUNK_REQD_ARGS = ["host"]
    _CONNECT_DEFAULTS: Dict[str, Any] = {"port": "8089"}
    _TIME_FORMAT = '"%Y-%m-%d %H:%M:%S.%6N"'

    def __init__(self, **kwargs):
        """Instantiate Splunk Driver."""
        super().__init__(**kwargs)
        self.service = None
        self._loaded = True
        self._connected = False
        if kwargs.get("debug", False):
            logger.setLevel(logging.DEBUG)
        self._required_params = self._SPLUNK_REQD_ARGS

        self.set_driver_property(
            DriverProps.PUBLIC_ATTRS,
            {
                "client": self.service,
                "saved_searches": self._saved_searches,
                "fired_alerts": self._fired_alerts,
            },
        )
        self.set_driver_property(
            DriverProps.FORMATTERS,
            {
                Formatters.DATETIME: self._format_datetime,
                Formatters.LIST: self._format_list,
            },
        )

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to Splunk via splunk-sdk.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Connection string with Splunk connection parameters

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/Splunk
        section of msticpyconfig.yaml, if available.

        """
        cs_dict = self._get_connect_args(connection_str, **kwargs)

        arg_dict = {
            key: val for key, val in cs_dict.items() if key in SPLUNK_CONNECT_ARGS
        }

        # Replace to Splunk python sdk's parameter name of sp_client.connect()
        if arg_dict.get("bearer_token"):
            arg_dict["splunkToken"] = arg_dict.pop("bearer_token")

        try:
            self.service = sp_client.connect(**arg_dict)
        except AuthenticationError as err:
            raise MsticpyConnectionError(
                f"Authentication error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        except HTTPError as err:
            raise MsticpyConnectionError(
                f"Communication error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        except Exception as err:
            raise MsticpyConnectionError(
                f"Error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        self._connected = True
        print("Connected.")

    def _get_connect_args(
        self, connection_str: Optional[str], **kwargs
    ) -> Dict[str, Any]:
        """Check and consolidate connection parameters."""
        cs_dict: Dict[str, Any] = self._CONNECT_DEFAULTS
        # Fetch any config settings
        cs_dict.update(self._get_config_settings("Splunk"))
        # If a connection string - parse this and add to config
        if connection_str:
            print("Credential loading from connection_str tied with ';'.")
            cs_items = connection_str.split(";")
            cs_dict.update(
                {
                    cs_item.split("=")[0].strip(): cs_item.split("=")[1]
                    for cs_item in cs_items
                }
            )
        elif kwargs:
            print("Credential loading from connect() method's args.")
            # if connection args supplied as kwargs
            cs_dict.update(kwargs)
            check_kwargs(cs_dict, list(SPLUNK_CONNECT_ARGS.keys()))
        else:
            print("Credential loading from msticpyconfig.yaml file.")

        cs_dict["port"] = int(cs_dict["port"])
        verify_opt = cs_dict.get("verify")
        if isinstance(verify_opt, str):
            cs_dict["verify"] = "true" in verify_opt.casefold()
        elif isinstance(verify_opt, bool):
            cs_dict["verify"] = verify_opt

        # Different required parameters for the REST API authentication method
        # between user/pass and authorization bearer token
        if "username" in cs_dict:
            self._required_params = ["host", "username", "password"]
            print("username/password method is selected.")
        else:
            self._required_params = ["host", "bearer_token"]
            print("bearer_token method is selected.")

        missing_args = set(self._required_params) - cs_dict.keys()
        if missing_args:
            raise MsticpyUserConfigError(
                "One or more connection parameters missing for Splunk connector",
                ", ".join(missing_args),
                f"Required parameters are {', '.join(self._required_params)}",
                "All parameters:",
                *[f"{arg}: {desc}" for arg, desc in SPLUNK_CONNECT_ARGS.items()],
                title="no Splunk connection parameters",
            )

        if "bearer_token" in self._required_params:
            epoch_now = datetime.now().timestamp()
            alg = jwt.get_unverified_header(cs_dict["bearer_token"])["alg"]
            decoded_token = jwt.decode(
                cs_dict["bearer_token"],
                algorithms=[alg],
                options={"verify_signature": False},
            )
            token_idp = decoded_token.get("idp")

            if token_idp == "Splunk":  # nosec
                epoch_issued_at = decoded_token["iat"]
                epoch_expiration = decoded_token["exp"]
                token_exp = datetime.fromtimestamp(epoch_expiration, timezone.utc)
                if epoch_issued_at < epoch_now < epoch_expiration:
                    print(f"This bearer_token is active until {token_exp}")
                else:
                    raise MsticpyConnectionError(
                        f"This bearer_token has been expired on {token_exp}",
                        "Recreate new bearer_token in your target Splunk and set it on.",
                        title="expired Splunk auth token",
                    )
        return cs_dict

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute splunk query and retrieve results via OneShot or async search mode.

        Parameters
        ----------
        query : str
            Splunk query to execute via OneShot or async search mode
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        count : int, optional
            Passed to Splunk job that indicates the maximum number
            of entities to return. A value of 0 indicates no maximum,
            by default, 0
        oneshot : bool, optional
            Set to True for oneshot (blocking) mode, by default False
        page_size = int, optional
            Pass to Splunk results reader in terms of fetch speed,
            which sets of result amount will be got at a time,
            by default, 100
        timeout : int, optional
            Amount of time to wait for results, by default 60

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err("Splunk")

        # default to unlimited query unless count is specified
        count = kwargs.pop("count", 0)

        # Get sets of N results at a time, N=100 by default
        page_size = kwargs.pop("page_size", 100)

        # default to 60s unless timeout is specified
        timeout = kwargs.pop("timeout", 60)

        # Normal (non-blocking) searches or oneshot (blocking) searches.
        # Defaults to Normal(non-blocking)

        # Oneshot is a blocking search that is scheduled to run immediately.
        # Instead of returning a search job, this mode returns the results
        # of the search once completed.
        # Because this is a blocking search, the results are not available
        # until the search has finished.
        # https://dev.splunk.com/enterprise/docs/python/
        # sdk-python/howtousesplunkpython/howtorunsearchespython
        is_oneshot = kwargs.get("oneshot", False)

        if is_oneshot is True:
            kwargs["output_mode"] = "json"
            query_results = self.service.jobs.oneshot(query, count=count, **kwargs)

            reader = sp_results.JSONResultsReader(  # pylint: disable=no-member
                query_results
            )  # due to DeprecationWarning of normal ResultsReader
            resp_rows = [row for row in reader if isinstance(row, dict)]
        else:
            # Set mode and initialize async job
            kwargs_normalsearch = {"exec_mode": "normal"}
            query_job = self.service.jobs.create(
                query, count=count, **kwargs_normalsearch
            )
            resp_rows, reader = self._exec_async_search(
                query_job, page_size, timeout=timeout
            )

        if len(resp_rows) == 0 or not resp_rows:
            print("Warning - query did not return any results.")
            return [row for row in reader if isinstance(row, sp_results.Message)]
        return pd.DataFrame(resp_rows)

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute against splunk instance.

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """
        raise NotImplementedError(f"Not supported for {self.__class__.__name__}")

    @property
    def service_queries(self) -> Tuple[Dict[str, str], str]:
        """
        Return dynamic queries available on connection to service.

        Returns
        -------
        Tuple[Dict[str, str], str]
            Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        if not self.connected:
            raise self._create_not_connected_err("Splunk")
        if hasattr(self.service, "saved_searches") and self.service.saved_searches:
            queries = {
                search.name.strip().replace(" ", "_"): f"search {search['search']}"
                for search in self.service.saved_searches
            }
            return queries, "SavedSearches"
        return {}, "SavedSearches"

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return dynamic queries available on connection to service.

        Returns
        -------
        Iterable[Dict[str, Any]]
            List of queries with properties: "name", "query", "container"
            and (optionally) "description"

        Raises
        ------
        MsticpyNotConnectedError
            If called before driver is connected.

        """
        if not self.connected:
            raise self._create_not_connected_err("Splunk")
        if hasattr(self.service, "saved_searches") and self.service.saved_searches:
            return [
                {
                    "name": search.name.strip().replace(" ", "_"),
                    "query": f"search {search['search']}",
                    "query_paths": "SavedSearches",
                    "description": "",
                }
                for search in self.service.saved_searches
            ]
        return []

    def _exec_async_search(self, query_job, page_size, timeout):
        """Execute an async search and return results."""
        # Initiate progress bar and start while loop, waiting for async query to complete
        progress_bar = tqdm(total=100, desc="Waiting Splunk job to complete")
        prev_progress = 0
        offset = 0  # Start at result 0
        start_time = datetime.now()
        end_time = start_time + timedelta(seconds=timeout)
        while True:
            while not query_job.is_ready():
                sleep(1)
            if self._retrieve_job_status(query_job, progress_bar, prev_progress):
                break
            if datetime.now() > end_time:
                raise MsticpyDataQueryError(
                    "Timeout waiting for Splunk query to complete",
                    f"Job completion reported {query_job['doneProgress']}",
                    title="Splunk query timeout",
                )
            sleep(1)
        # Update progress bar indicating job completion
        progress_bar.update(100)
        progress_bar.close()
        sleep(2)

        logger.info("Implicit parameter dump - 'page_size': %d", page_size)
        return self._retrieve_results(query_job, offset, page_size)

    @staticmethod
    def _retrieve_job_status(query_job, progress_bar, prev_progress):
        """Poll the status of a job and update the progress bar."""
        stats = {
            "is_done": query_job["isDone"],
            "done_progress": float(query_job["doneProgress"]) * 100,
            "scan_count": int(query_job["scanCount"]),
            "event_count": int(query_job["eventCount"]),
            "result_count": int(query_job["resultCount"]),
        }
        status = (
            "\r%(done_progress)03.1f%%   %(scan_count)d scanned   "
            "%(event_count)d matched   %(result_count)d results"
        ) % stats
        if prev_progress == 0:
            progress = stats["done_progress"]
        else:
            progress = stats["done_progress"] - prev_progress
        prev_progress = stats["done_progress"]
        progress_bar.update(progress)

        if stats["is_done"] == "1":
            logger.info(status)
            logger.info("Splunk job completed.")
            return True
        return False

    @staticmethod
    def _retrieve_results(query_job, offset, page_size):
        """Retrieve the results of a job, decode and return them."""
        # Retrieving all the results by paginate
        result_count = int(
            query_job["resultCount"]
        )  # Number of results this job returned

        resp_rows = []
        progress_bar_paginate = tqdm(
            total=result_count, desc="Waiting Splunk result to retrieve"
        )
        while offset < result_count:
            kwargs_paginate = {
                "count": page_size,
                "offset": offset,
                "output_mode": "json",
            }
            # Get the search results and display them
            search_results = query_job.results(**kwargs_paginate)
            # due to DeprecationWarning of normal ResultsReader
            reader = sp_results.JSONResultsReader(  # pylint: disable=no-member
                search_results
            )
            resp_rows.extend([row for row in reader if isinstance(row, dict)])
            progress_bar_paginate.update(page_size)
            offset += page_size
        # Update progress bar indicating fetch results
        progress_bar_paginate.update(result_count)
        progress_bar_paginate.close()
        logger.info("Retrieved %d results.", len(resp_rows))
        return resp_rows, reader

    @property
    def _saved_searches(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of saved searches in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of saved searches with name and query columns.

        """
        return self._get_saved_searches() if self.connected else None

    def _get_saved_searches(self) -> Union[pd.DataFrame, Any]:
        # sourcery skip: class-extract-method
        """
        Return list of saved searches in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of saved searches with name and query columns.

        """
        if not self.connected:
            raise self._create_not_connected_err("Splunk")
        savedsearches = self.service.saved_searches

        out_df = pd.DataFrame(columns=["name", "query"])

        namelist = []
        querylist = []
        for savedsearch in savedsearches:
            namelist.append(savedsearch.name.replace(" ", "_"))
            querylist.append(savedsearch["search"])
        out_df["name"] = namelist
        out_df["query"] = querylist

        return out_df

    @property
    def _fired_alerts(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of fired alerts in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of fired alerts with alert name and count columns.

        """
        return self._get_fired_alerts() if self.connected else None

    def _get_fired_alerts(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of fired alerts in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of fired alerts with alert name and count columns.

        """
        if not self.connected:
            raise self._create_not_connected_err("Splunk")
        firedalerts = self.service.fired_alerts

        out_df = pd.DataFrame(columns=["name", "count"])

        alert_names = []
        alert_counts = []
        for alert in firedalerts:
            alert_names.append(alert.name)
            alert_counts.append(alert.count)
        out_df["name"] = alert_names
        out_df["count"] = alert_counts

        return out_df

    # Parameter Formatting methods
    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return f'"{date_time.isoformat(sep=" ")}"'

    @staticmethod
    def _format_list(param_list: Iterable[Any]) -> str:
        """Return formatted list parameter."""
        fmt_list = [f'"{item}"' for item in param_list]
        return ",".join(fmt_list)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
nbtools module - Notebook Security Tools.

This is a collection of modules with functionality (mostly) specific to
notebooks. It also houses some visualization modules that will migrate
to the vis sub-package.

- nbinit - notebook initialization
- azure_ml_tools - configuration and helpers for AML workspaces
- nbwidgets - ipywidgets-based UI components for infosec notebooks
- nbdisplay - miscellaneous display functions TBM to vis

"""
# flake8: noqa: F403
# pylint: disable=W0401
# import importlib
# from typing import Any

# from .. import nbwidgets
from .._version import VERSION
from ..lazy_importer import lazy_import

# from ..common import utility as utils
# from ..common.wsconfig import WorkspaceConfig
# from ..vis import nbdisplay
# from .security_alert import SecurityAlert

# try:
#     from IPython import get_ipython

#     from ..init import nbmagics
# except ImportError as err:
#     pass

# pylint: enable=W0401

__version__ = VERSION

# _DEFAULT_IMPORTS = {"nbinit": "msticpy.init.nbinit"}

_LAZY_IMPORTS = {
    "msticpy.init.nbinit",
    "msticpy.common.utility as utils",
    "msticpy.common.wsconfig.WorkspaceConfig",
    "msticpy.nbtools.security_alert.SecurityAlert",
    "msticpy.nbwidgets",
    "msticpy.vis.nbdisplay",
}

# def __getattr__(attrib: str) -> Any:
#     """
#     Import and return an attribute of nbtools.

#     Parameters
#     ----------
#     attrib : str
#         The attribute name

#     Returns
#     -------
#     Any
#         The attribute value.

#     Raises
#     ------
#     AttributeError
#         No attribute found.

#     """
#     if attrib in _DEFAULT_IMPORTS:
#         module = importlib.import_module(_DEFAULT_IMPORTS[attrib])
#         return module
#     raise AttributeError(f"msticpy has no attribute {attrib}")

# from .vtlookupv3 import VT3_AVAILABLE

# vtlookupv3: Any
# if VT3_AVAILABLE:
#     from .vtlookupv3 import vtlookupv3
# else:
#     # vtlookup3 will not load if vt package not installed
#     vtlookupv3 = ImportPlaceholder(  # type: ignore
#         "vtlookupv3", ["vt-py", "vt-graph-api", "nest_asyncio"]
#     )

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)
